import contextlib
import logging
import sys
import warnings

from pandas import options

verbose = True  # Controls the amount of feedback from the I-Server
fetch_on_init = True  # Controls if object will fetch basic data from server on init
progress_bar = (
    True  # Controls whether progress bar will be shown during long fetch operations
)
debug = False  # Lets the program run in debugging mode
# Sets number of rows displayed for pandas DataFrame
options.display.max_rows = max(250, options.display.max_rows)
options.display.max_colwidth = max(100, options.display.max_colwidth)
# Warning settings: "error", "ignore", "always", "default", "module", "once"
print_warnings = 'always'
module_path = 'mstrio.*'
save_responses = False  # Used to save REST API responses for mocking
wip_warnings_enabled = (
    True  # Controls whether warnings/errors about WIP functionality are emitted
)


def custom_formatwarning(msg, category, *args, **kwargs):
    # ignore everything except the message
    return str(category.__name__) + ': ' + str(msg) + '\n'


warnings.formatwarning = custom_formatwarning
warnings.filterwarnings(action=print_warnings, module=module_path)
warnings.filterwarnings(
    action=print_warnings, category=DeprecationWarning, module=module_path
)
warnings.filterwarnings(action='default', category=UserWarning, module=module_path)


def get_logging_level() -> int:
    """Calculate and return logging level
    to configure logger.
    """
    return logging.DEBUG if debug else logging.INFO


logger_stream_handler = logging.StreamHandler(stream=sys.stdout)

# warns issued by the warnings module will be redirected to the logging.warning
logging.captureWarnings(True)

# root logger for mstrio, all loggers in submodules will have keys of the form
# "mstrio.sub1.sub2" and have this one as its ancestor
logger = logging.getLogger("mstrio")
logger.propagate = False
warnings_logger = logging.getLogger("py.warnings")

logger.addHandler(logger_stream_handler)
logger.setLevel(get_logging_level())


def toggle_debug_mode() -> None:
    """Toggle debug mode between INFO and DEBUG.
    It will change root logger's logging level.
    """
    global debug
    debug = not debug
    logger.setLevel(get_logging_level())


@contextlib.contextmanager
def temp_verbose_disable():
    """Temporarily disable verbose logging, with keeping current state.

    Built to be used as context with `with` statement:

        >>> with temp_verbose_disable():
        >>>     ...
    """
    global verbose
    old_verbose = verbose
    verbose = False
    try:
        yield
    finally:
        verbose = old_verbose



"""List names and IDs of empty user groups.

1. Connect to the environment using data from workstation
2. Get all user groups
3. Filter out user groups that have no members
4. Print names and ids of filtered user groups
"""

from typing import List

from mstrio.connection import Connection, get_connection
from mstrio.users_and_groups import list_user_groups, UserGroup


def list_empty_user_groups(connection: "Connection") -> List["UserGroup"]:
    """List empty user groups.

    Args:
        connection: Strategy One connection object returned by
            `connection.Connection()`
    """
    all_user_groups = list_user_groups(connection=connection)
    return [
        user_group_ for user_group_ in all_user_groups if not user_group_.list_members()
    ]


# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, 'MicroStrategy Tutorial')

# get empty user groups
empty_user_groups = list_empty_user_groups(conn)

# print empty user groups
print('Empty user groups:', flush=True)
for ug in empty_user_groups:
    print(f"{ug.name} ({ug.id}) ", flush=True)



"""This is the demo script to show how administrator can manage security
filters.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import get_connection
from mstrio.modeling import (
    Attribute,
    AttributeFormPredicate,
    ConstantParameter,
    Expression,
    Function,
    list_attributes,
    ObjectSubType,
    Operator,
    SchemaObjectReference,
    Variant,
    VariantType
)
from mstrio.modeling.security_filter import list_security_filters, SecurityFilter
from mstrio.users_and_groups import list_user_groups, list_users

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Project to connect to

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# List all security filters
security_filters = list_security_filters(conn)
print(security_filters)

# Define variables which can be later used in a script
SECURITY_FILTER_ID = $security_filter_id
SECURITY_FILTER_NAME = $security_filter_name

# Get a security filter by id
security_filter = SecurityFilter(conn, id=SECURITY_FILTER_ID)
print(security_filter)

# Get a security filter by name
security_filter = SecurityFilter(conn, name=SECURITY_FILTER_NAME)
print(security_filter)

# List the properties of the security filter
print(security_filter.list_properties())

# Show the qualification of the security filter
print(security_filter.qualification)

# List the members of the security filter
print(security_filter.members)

# Define variables which can be later used in a script
NEW_SECURITY_FILTER_NAME = $new_security_filter_name
CONSTANT_VALUE_1 = $constant_value_1
ATTRIBUTE_ID = $attribute_id
ATTRIBUTE_NAME = $attribute_name
ATTRIBUTE_FORM_ID = $attribute_form_id
ATTRIBUTE_FORM_NAME = $attribute_form_name

# create new security filter
QUALIFICATION = Expression(
    tree=AttributeFormPredicate(
        # see access_and_security/security_filter/predicates.py
        # for PredicateFormFunction values
        function=Function.GREATER,
        # see access_and_security/security_filter/predicate_parameters.py
        # for ConstantType values
        parameters=[
            ConstantParameter(constant=Variant(type=VariantType.STRING, value=CONSTANT_VALUE_1))
        ],
        # insert object id and its attribute name to which you want to refer to
        attribute=SchemaObjectReference(
            sub_type=ObjectSubType.ATTRIBUTE, object_id=ATTRIBUTE_ID, name=ATTRIBUTE_NAME
        ),
        # insert object id and its attribute form system name
        # to which you want to refer to
        form=SchemaObjectReference(
            sub_type=ObjectSubType.ATTRIBUTE_FORM_SYSTEM,
            object_id=ATTRIBUTE_FORM_ID,
            name=ATTRIBUTE_FORM_NAME
        )
    )
)

# Define a variable which can be later used in a script
FOLDER_ID = $folder_id  # folder in which new security filter will be created

new_security_filter = SecurityFilter.create(
    connection=conn,
    qualification=QUALIFICATION,
    name=NEW_SECURITY_FILTER_NAME,
    destination_folder=FOLDER_ID
)
print(new_security_filter)


# Create new security filter providing string as qualification expression.
NEW_SECURITY_FILTER_NAME_STRING = $new_security_filter_name_string
QUALIFICATION_STRING = '(Year(CurrentDate()) - 5)< Year@ID'

new_security_filter_string = SecurityFilter.create(
    connection=conn,
    qualification=QUALIFICATION_STRING,
    name=NEW_SECURITY_FILTER_NAME_STRING,
    destination_folder=FOLDER_ID
)
print(new_security_filter_string)

# Create new security filter with more than one attribute form.

# 1. List attributes to get attribute IDs
attributes = list_attributes(conn)
print(attributes)

# Define variables which can be later used in a script
ATTRIBUTE_1_ID = $attribute_1_id
ATTRIBUTE_1_NAME = $attribute_1_name
ATTRIBUTE_2_ID = $attribute_2_id
ATTRIBUTE_2_NAME = $attribute_2_name

# 2. Find attribute forms for the particular attribute
attribute_1_forms = Attribute(conn, id=ATTRIBUTE_1_ID).forms
print(attribute_1_forms)

attribute_2_forms = Attribute(conn, id=ATTRIBUTE_2_ID).forms
print(attribute_2_forms)

# 3. To build security filter qualification where there are two attribute
# forms you have to create two separate `AttributeFormPredicate` objects
# and connect them with logic function using one value of `Function` enum.
# You cannot set values for two attribute forms in a single
# `AttributeFormPredicate`.

# Define variables which can be later used in a script
ATTRIBUTE_FORM_1_ID = $attribute_form_1_id
ATTRIBUTE_FORM_2_ID = $attribute_form_2_id
CONSTANT_VALUE_2 = $constant_value_2
CONSTANT_VALUE_3 = $constant_value_3
ATTRIBUTE_FORM_1_ID = $attribute_form_1_id
ATTRIBUTE_FORM_2_ID = $attribute_form_2_id

predicate1 = AttributeFormPredicate(
    function=Function.EQUALS,
    parameters=[
        ConstantParameter(constant=Variant(type=VariantType.STRING, value=CONSTANT_VALUE_2))
    ],
    attribute=SchemaObjectReference(
        sub_type=ObjectSubType.ATTRIBUTE, object_id=ATTRIBUTE_1_ID, name=ATTRIBUTE_1_NAME
    ),
    form=SchemaObjectReference(
        sub_type=ObjectSubType.ATTRIBUTE_FORM_SYSTEM,
        object_id=ATTRIBUTE_FORM_1_ID,
    ),
)
predicate2 = AttributeFormPredicate(
    function=Function.EQUALS,
    parameters=[
        ConstantParameter(constant=Variant(type=VariantType.STRING, value=CONSTANT_VALUE_2))
    ],
    attribute=SchemaObjectReference(
        sub_type=ObjectSubType.ATTRIBUTE, object_id=ATTRIBUTE_2_ID, name=ATTRIBUTE_2_NAME
    ),
    form=SchemaObjectReference(
        sub_type=ObjectSubType.ATTRIBUTE_FORM_SYSTEM,
        object_id=ATTRIBUTE_FORM_2_ID,
    ),
)

# Define variables which can be later used in a script
NEW_SECURITY_FILTER_NAME_2 = $new_security_filter_name_2
NEW_SECURITY_FILTER_DESCRIPTION = $new_security_filter_description

new_security_filter2 = SecurityFilter.create(
    connection=conn,
    qualification=Expression(
        tree=Operator(function=Function.AND, children=[predicate1, predicate2])
    ),
    name=NEW_SECURITY_FILTER_NAME_2,
    destination_folder=FOLDER_ID
)
print(new_security_filter2)

# alter security filter - its name, description, qualification and folder
security_filter.alter(name=NEW_SECURITY_FILTER_NAME, description=NEW_SECURITY_FILTER_DESCRIPTION)
print(security_filter)

# Define a variable which can be later used in a script
PARAMETER_CHANGED_VALUE = $parameter_changed_value  # value of parameter to change to

# copy old qualification and change value of parameter
new_qualification = security_filter.qualification
new_qualification.tree.parameters[0].constant.value = PARAMETER_CHANGED_VALUE

security_filter.alter(qualification=new_qualification)
print(security_filter)

# Define variables which can be later used in a script
USERNAME_BEGINS = $username_begins  # beginning of username to look for
USER_GROUP_NAME_BEGINS = $user_group_name_begins  # beginning of user group to look for

# prepare users and user groups for applying and revoking security filter
user = list_users(conn, name_begins=USERNAME_BEGINS)[0]
group = list_user_groups(conn, name_begins=USER_GROUP_NAME_BEGINS)[0]

# apply user(s) and/or user group(s) to security filter
security_filter.apply([user, group.id])

# revoke user(s) and/or user group(s) from security filter
security_filter.revoke([user.id, group])

# revoke all members of security filter
security_filter.revoke(security_filter.members)

# apply security filter directly to user or user group
user.apply_security_filter(security_filter)

# revoke security filter directly from user or user group
group.apply_security_filter(security_filter.id)

# fetch the latest security filter state
security_filter.fetch()

# delete security filter. When `force` argument is set
# to `False` (default value) then deletion must be approved.
security_filter.delete(force=True)



import logging
from collections import deque
from typing import TYPE_CHECKING

from mstrio import config
from mstrio.api import folders
from mstrio.object_management import PredefinedFolders
from mstrio.types import ObjectTypes
from mstrio.users_and_groups import User
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity, MoveMixin
from mstrio.utils.helper import (
    fetch_objects_async,
    get_default_args_from_func,
    get_temp_connection,
)
from mstrio.utils.resolvers import (
    get_project_id_from_params_set,
    validate_owner_key_in_filters,
)
from mstrio.utils.response_processors import objects as objects_processors

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.server.project import Project

logger = logging.getLogger(__name__)


def list_folders(
    connection: "Connection",
    project: "Project | str | None" = None,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    include_subfolders: bool = False,
    parent_folder: "str | Folder | None" = None,
    **filters,
) -> list["Folder"] | list[dict]:
    """Get a list of folders - either all folders in a specific project or all
    folders that are outside of projects, called configuration-level folders.
    The list of configuration-level folders includes folders such as users, user
    groups, databases, etc. which are not project-specific. If you pass
    any project-related parameter, you get folders in that project;
    if not, then you get configuration-level folders.

    Note:
        Id of project is not taken directly from `Connection` object, so you
        have to specify it explicitly.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        to_dictionary (bool, optional): If True returns dicts, by default
            (False) returns objects.
        limit (int): limit the number of elements returned. If `None` (default),
            all objects are returned.
        include_subfolders (bool): if True, all subfolders are included in the
            result. Default is False.
            Note that using this option will result in a large number of
                objects being fetched, as it will fetch all folders available.
        parent_folder (string or Folder, optional): Parent folder from which to
            list folders. Can be provided with three forms of input:
                - Folder object
                - Folder ID
                - Folder path. The path has to be provided in the following
                    format:
                        if it's inside of a project, example:
                            /MicroStrategy Tutorial/Public Objects/Metrics
                        if it's a root folder, example:
                            /CASTOR_SERVER_CONFIGURATION/Users

        **filters: Available filter parameters: ['name', 'id', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'acg',
             'owner', 'hidden',
          'ext_type']

    Returns:
        list of `Folder` objects or list of dictionaries
    """
    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
        assert_id_exists=False,
        no_fallback_from_connection=True,
    )

    validate_owner_key_in_filters(filters)

    # Project is validated only if project was specified in arguments -
    # otherwise fetch is performed from a non-project area.
    temp_conn = get_temp_connection(connection, proj_id)

    if not parent_folder:
        objects = fetch_objects_async(
            temp_conn,
            folders.list_folders,
            folders.list_folders_async,
            limit=limit,
            chunk_size=1000,
            project_id=proj_id,
            filters=filters,
        )
    else:
        parent_folder_id = _get_parent_folder_id(temp_conn, parent_folder)

        with config.temp_verbose_disable():
            parent_folder = Folder(connection=temp_conn, id=parent_folder_id)
            objects = parent_folder.get_contents(
                include_subfolders=include_subfolders,
                limit=limit,
                to_dictionary=True,
                type=8,
            )

    if include_subfolders:
        recursive_objects = []

        with config.temp_verbose_disable():
            new_limit = limit - len(objects) if limit else None
            for obj in objects:
                folder = Folder(connection=temp_conn, id=obj.get('id'))
                recursive_objects += folder.get_contents(
                    to_dictionary=True, include_subfolders=True, limit=new_limit, type=8
                )

        objects += recursive_objects

    objects = objects[:limit]

    if to_dictionary:
        return objects
    else:
        from mstrio.utils.object_mapping import map_objects_list

        return map_objects_list(temp_conn, objects)


def get_my_personal_objects_contents(
    connection: "Connection",
    project: "Project | str | None" = None,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
) -> list:
    """Get contents of `My Personal Objects` folder in a specific project.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        to_dictionary (bool, optional): If True returns dicts, by default
            (False) returns objects.

    Returns:
        list of objects or list of dictionaries
    """
    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
    )

    objects = folders.get_my_personal_objects_contents(connection, proj_id).json()
    if to_dictionary:
        return objects
    else:
        from mstrio.utils.object_mapping import map_objects_list

        return map_objects_list(connection, objects)


def get_predefined_folder_contents(
    connection: "Connection",
    folder_type: PredefinedFolders,
    project: "Project | str | None" = None,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list:
    """Get contents of a pre-defined Strategy One folder in a specific project.
    Available values for `folder_type` are stored in enum `PredefinedFolders`.

    Args:
        connection (object): Strategy One connection object returned by
            `connection.Connection()`
        folder_type (enum): pre-defined folder type. Available values are
            stored in enum `PredefinedFolders`.
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        to_dictionary (bool, optional): If True returns dicts, by default
            (False) returns objects.
        limit (int): limit the number of elements returned. If `None` (default),
            all objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'date_created', 'date_modified', 'acg']

    Returns:
        list of objects or list of dictionaries
    """

    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
    )

    objects = fetch_objects_async(
        connection,
        folders.get_predefined_folder_contents,
        folders.get_predefined_folder_contents_async,
        limit=limit,
        chunk_size=1000,
        folder_type=folder_type.value,
        project_id=proj_id,
        filters=filters,
    )

    if to_dictionary:
        return objects
    else:
        from mstrio.utils.object_mapping import map_objects_list

        return map_objects_list(connection, objects)


def get_folder_id_from_path(connection: "Connection", path: str) -> str:
    """Get folder id from folder path.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        path (str): path of the Folder
            the path has to be provided in the following format:
                if it's inside of a project, example:
                    /MicroStrategy Tutorial/Public Objects/Metrics
                if it's a root folder, example:
                    /CASTOR_SERVER_CONFIGURATION/Users

    Returns:
        Folder id.
    """
    if path[0] == '/':
        path = path[1:]
    if path[-1] == '/':
        path = path[:-1]
    folders_in_path = path.split('/')
    project_id = (
        # FYI: EXPECT project ID if project-level...
        get_project_id_from_params_set(connection, project_name=folders_in_path[0])
        if folders_in_path[0] != 'CASTOR_SERVER_CONFIGURATION'
        # ... `None` only otherwise
        else None
    )
    try:
        if project_id:
            original_project_id = connection.project_id
            connection.project_id = project_id
            # This root Project folder is hardcoded in every Environment
            folder = Folder(
                connection=connection, id='D43364C684E34A5F9B2F9AD7108F7828'
            )
        else:
            castor_id = (
                folders.get_predefined_folder_id(connection=connection, folder_type=39)
                .json()
                .get('id')
            )
            folder = Folder(connection=connection, id=castor_id)
        for i in range(1, len(folders_in_path)):
            temp_ids = [
                item.get('id')
                for item in folder.get_contents(to_dictionary=True)
                if item.get('type') == 8 and item.get('name') == folders_in_path[i]
            ]
            if not temp_ids:
                error_message = (
                    f"Couldn't find folder with given name {folders_in_path[i]} "
                    f"while exploring the path. Check if provided path is correct."
                )
                raise ValueError(error_message)
            folder = Folder(
                connection=connection,
                id=temp_ids[0],
            )
        return folder.id
    finally:
        if project_id:
            connection.project_id = (
                original_project_id if original_project_id else connection.project_id
            )


def _get_parent_folder_id(
    temp_conn: "Connection",
    parent_folder: "str | Folder | None",
) -> str:
    if isinstance(parent_folder, str) and '/' in parent_folder:
        with config.temp_verbose_disable():
            return get_folder_id_from_path(temp_conn, parent_folder)
    elif isinstance(parent_folder, Folder):
        return parent_folder.id
    else:
        return parent_folder


class Folder(Entity, CopyMixin, MoveMixin, DeleteMixin):
    """Object representation of Strategy One Folder object.

    Attributes:
        connection: MSTR Connection object
        id: ID of the folder
        name: name of the folder
        description: description of the folder
        type: the type of the folder (based on EnumDSSXMLObjectTypes)
        subtype: the subtype of the folder (based on EnumDSSXMLObjectSubTypes)
        date_created: the date/time at which the folder was first saved into the
            metadata
        date_modified: the date/time at which the folder was last saved into the
            metadata
        version: the version number this folder is currently carrying
        acl: an array of access control entry objects
        owner: User object, the owner of the object
        acg: access rights, EnumDSSXMLAccessRightFlags
        ext_type: object extended type, EnumDSSExtendedType
        ancestors: list of ancestor folders
        abbreviation: folder's abbreviation
        hidden: specifies whether the folder is hidden
        icon_path: folder icon path
        view_media: view media settings
        certified_info: CertifiedInfo object, certification status, time of
            certification, and information about the certifier (currently only
            for document and report)
        contents: contents of folder
    """

    _FROM_DICT_MAP = {**Entity._FROM_DICT_MAP, 'owner': User.from_dict}

    _API_PATCH: dict = {
        (
            'name',
            'description',
            'abbreviation',
            'hidden',
            'folder_id',
            'comments',
            'owner',
        ): (objects_processors.update, 'partial_put')
    }

    _OBJECT_TYPE = ObjectTypes.FOLDER
    _SIZE_LIMIT = 10000000  # this sets desired chunk size in bytes

    def __init__(
        self,
        connection: "Connection",
        id: str | None = None,
        path: str | None = None,
        name: str | None = None,
    ):
        """Initialize folder object by its identifier.

        Note:
            Parameter `name` is not used when fetching. `id` is always used to
            uniquely identify folder.

        Args:
            connection: Strategy One connection object returned by
                `connection.Connection()`.
            id (str, optional): Identifier of a pre-existing folder containing
                the required data.
            path (str, optional): path of a a pre-existing folder containing
                the required data. Can be provided as an alternative to
                `id` parameter. If both are provided, `id` is used.
                    the path has to be provided in the following format:
                        if it's inside of a project, example:
                            /MicroStrategy Tutorial/Public Objects/Metrics
                        if it's a root folder, example:
                            /CASTOR_SERVER_CONFIGURATION/Users
            name (str): name of folder.
        """
        if not id:
            if path:
                id = get_folder_id_from_path(connection, path)
            else:
                raise ValueError("Either 'id' or 'path' has to be provided.")

        super().__init__(connection, id, name=name)

    @classmethod
    def create(
        cls,
        connection: "Connection",
        name: str,
        parent: str | None = None,
        parent_path: str | None = None,
        description: str | None = None,
    ) -> "Folder":
        """Create a new folder in a folder selected within connection object
        by providing its name, id of parent folder and optionally description.

        Args:
            connection: Strategy One connection object returned by
                `connection.Connection()`.
            name (str): name of a new folder.
            parent (str): id of a parent folder in which new folder will be
                created.
            parent_path (str, optional): path of a parent folder in which new
                folder will be created. Can be provided as an alternative to
                `parent` parameter. If both are provided, `parent` is used.
                    the path has to be provided in the following format:
                        if it's inside of a project, example:
                            /MicroStrategy Tutorial/Public Objects/Metrics
                        if it's a root folder, example:
                            /CASTOR_SERVER_CONFIGURATION/Users
            description (str, optional): optional description of a new folder.

        Returns:
            newly created folder
        """
        connection._validate_project_selected()

        if not parent:
            if parent_path:
                parent = get_folder_id_from_path(connection, parent_path)
            else:
                raise ValueError("Either 'parent' or 'parent_path' has to be provided.")

        response = folders.create_folder(connection, name, parent, description).json()
        if config.verbose:
            logger.info(
                f"Successfully created folder named: '{response.get('name')}' "
                f"with ID: '{response.get('id')}'"
            )
        return cls.from_dict(source=response, connection=connection)

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        hidden: bool | None = None,
        comments: str | None = None,
        owner: str | User | None = None,
    ) -> None:
        """Alter the folder properties.

        Args:
            name: folder name
            description: folder description
            hidden: Specifies whether the folder is hidden
            comments: long description of the folder
            owner: (str, User, optional): owner of the folder
        """
        if isinstance(owner, User):
            owner = owner.id
        func = self.alter
        default_dict = get_default_args_from_func(func)
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    def get_contents(
        self,
        to_dictionary: bool = False,
        include_subfolders=False,
        limit: int | None = None,
        **filters,
    ) -> list:
        """Get contents of a folder. It can contain other folders or different
        kinds of objects.

        Args:
            to_dictionary (bool, optional): If True returns dicts, by default
                (False) returns objects.
            include_subfolders (bool, optional): If True returns contents of all
                subfolders as well. False by default.
                Note that using this option may result in a large number of
                objects being fetched, especially if coupled with filters.
            limit (int, optional): Limit the number of elements returned. If
                `None` (default), all objects are returned.
            **filters: Available filter parameters: ['name', 'id', 'type',
                'subtype', 'date_created', 'date_modified', 'version',
                'acg', 'owner', 'ext_type']

        Returns:
            Contents as Python objects (when `to_dictionary` is `False` (default
            value)) or contents as dictionaries otherwise.
        """
        queue = deque([self.id])
        objects = []

        validate_owner_key_in_filters(filters)

        while queue and (limit is None or len(objects) < limit):
            current_id = queue.popleft()
            current_objects = fetch_objects_async(
                self.connection,
                folders.get_folder_contents,
                folders.get_folder_contents_async,
                limit=limit,
                chunk_size=1000,
                id=current_id,
                filters=filters,
            )
            objects.extend(current_objects)

            if include_subfolders and (limit is None or len(objects) < limit):
                current_objects = fetch_objects_async(
                    self.connection,
                    folders.get_folder_contents,
                    folders.get_folder_contents_async,
                    limit=limit,
                    chunk_size=1000,
                    id=current_id,
                    filters={'type': 8},
                )
                child_folders = [
                    child for child in current_objects if child.get('type') == 8
                ]
                queue.extend(child.get('id') for child in child_folders)

        objects = objects[:limit]
        if to_dictionary:
            return objects
        else:
            from mstrio.utils.object_mapping import map_objects_list

            return map_objects_list(self.connection, objects)



"""This is the demo script to show how administrator can manage schema.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.helpers import IServerError
from mstrio.modeling.schema import SchemaManagement, SchemaLockType, SchemaUpdateType

from mstrio.connection import get_connection

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Project to connect to

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# create an object to manage the schema
schema_mgmt = SchemaManagement(conn)

# get lock type of the schema
lock_t = schema_mgmt.lock_type

# unlock schema and get its status
schema_mgmt.unlock()
lock_st = schema_mgmt.get_lock_status()

# Lock schema and get its status. Schema can be locked also with type
# `ABSOLUTE_CONSTITUENT` or `EXCLUSIVE_CONSTITUENT`.
# Available values for SchemaLockType enum are in modeling/schema/schema_management.py
schema_mgmt.lock(SchemaLockType.ABSOLUTE_INDIVIDUAL)
lock_st = schema_mgmt.get_lock_status()

# Reload schema. It is allowed not to provide an update type. Except for
# types below `TABLE_KEY` or `LOGICAL_SIZE` are also available.
# When schema is reloaded asynchronously then task is returned.
# Available values for SchemaLockType enum are in modeling/schema/schema_management.py
task = schema_mgmt.reload(
    update_types=[SchemaUpdateType.CLEAR_ELEMENT_CACHE, SchemaUpdateType.ENTRY_LEVEL]
)

# get list of tasks which are stored within schema management object (tasks
# are saved when `reload` is performed asynchronously)
tasks = schema_mgmt.tasks

# get all details about the first task which is stored within schema management
# object
task = schema_mgmt.get_task(task_index=0)

# get all details about the last task which is stored within schema management
# object
task_st = schema_mgmt.get_task(task_index=-1)

# Reload schema without asynchronous response. This will guarantee that the
# schema reload task is completed before the next operation is performed.
# If the task fails, an exception will be raised. This does not return any
# intermediate task details.
try:
    schema_mgmt.reload(respond_async=False)
except IServerError as e:
    print(f"Error occurred: {e}")

# unlock schema and get its status
schema_mgmt.unlock()
lock_st = schema_mgmt.get_lock_status()



import logging
from typing import TYPE_CHECKING

from mstrio import config
from mstrio.connection import Connection
from mstrio.types import ObjectTypes
from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from mstrio.utils.sessions import FuturesSessionWithRenewal

logger = logging.getLogger(__name__)


def _validate_project_id(
    source: 'Connection | FuturesSessionWithRenewal',
    object_type: int,
    project_id: str | None,
    log_info_on_scope: bool = True,
) -> str | None:
    is_asking_for_project = object_type == ObjectTypes.PROJECT.value
    if is_asking_for_project:
        project_id = None
    elif project_id is None:  # do NOT fall through if we want PROJECT
        project_id = (
            source.project_id
            if isinstance(source, Connection)
            else source.connection.project_id
        )

    if (
        log_info_on_scope
        and not is_asking_for_project
        and not project_id
        and config.verbose
    ):
        logger.info(
            'Project was not selected. Action is performed for the non-project area'
        )

    return project_id


@ErrorHandler(err_msg="Error getting information for the object with ID {id}")
def get_object_info(
    connection: Connection,
    id,
    object_type,
    project_id=None,
    error_msg=None,
    whitelist=[('ERR001 ', 500)],  # noqa: B006
):
    """Get information for a specific object in a specific project; if you do
    not specify a project ID, you get information for the object just in the
    non-project area.

    You identify the object with the object ID and object type. You specify
    the object type as a query parameter; possible values for object type are
    provided in EnumDSSXMLObjectTypes.

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        object_type (int): One of EnumDSSXMLObjectTypes. Ex. 34 (User or
        UserGroup), 44 (Security Role), 32 (Project), 8 (Folder), 36 (type of
        I-Server configuration), 58 (Security Filter)
        project_id(str): ID of a project in which the object is located.
        error_msg (string, optional): Custom Error Message for Error Handling
        whitelist (list, optional): List of tuples containing error codes and
            status codes that should be ignored by the error handler.
            Defaults to [('ERR001 ', 500)] - ignore "Object not found" errors.

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    project_id = _validate_project_id(connection, object_type, project_id)

    return connection.get(
        endpoint=f'/api/objects/{id}',
        headers={'X-MSTR-ProjectID': project_id},
        params={'type': object_type, 'comments': True},
    )


def get_object_info_async(
    future_session: 'FuturesSessionWithRenewal', id, object_type, project_id=None
):
    """Get information for a specific object in a specific project; if you do
    not specify a project ID, you get information for the object just in the
    non-project area.

    You identify the object with the object ID and object type. You specify
    the object type as a query parameter; possible values for object type are
    provided in EnumDSSXMLObjectTypes.

    Args:
        future_session (FuturesSessionWithRenewal): Future Session object
            to call Strategy One REST Server asynchronously
        id (str): Object ID
        object_type (int): One of EnumDSSXMLObjectTypes. Ex. 34 (User or
        UserGroup), 44 (Security Role), 32 (Project), 8 (Folder), 36 (type of
        I-Server configuration), 58 (Security Filter)
        project_id(str): ID of a project in which the object is located.

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    project_id = _validate_project_id(future_session, object_type, project_id)

    return future_session.get(
        endpoint=f'/api/objects/{id}',
        headers={'X-MSTR-ProjectID': project_id},
        params={'type': object_type},
    )


@ErrorHandler(err_msg="Error deleting object with ID {id}")
def delete_object(
    connection: Connection,
    id,
    object_type,
    project_id=None,
    error_msg=None,
):
    """Delete a specific object in a specific project; if you do not specify a
    project ID, you delete information for the object in all projects.

    You identify the object with the object ID and object type. You specify
    the object type as a query parameter; possible values for object type are
    provided in EnumDSSXMLObjectTypes.

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        object_type (int): One of EnumDSSXMLObjectTypes. Ex. 34 (User or
        UserGroup), 44 (Security Role), 32 (Project), 8 (Folder), 36 (type of
        I-Server configuration)
        project_id (str): ID of a project in which the object is located.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    project_id = _validate_project_id(connection, object_type, project_id, False)

    return connection.delete(
        endpoint=f'/api/objects/{id}',
        headers={'X-MSTR-ProjectID': project_id},
        params={'type': object_type},
    )


@ErrorHandler(err_msg="Error bulk deleting objects with IDs {ids}")
def bulk_delete_objects(
    connection: Connection,
    ids,
    object_types,
    project_id=None,
    error_msg=None,
):
    """Delete multiple objects in a specific project.

    You identify the objects with their object IDs and object types. You
    specify the object types as a query parameter; possible values for object
    type are provided in EnumDSSXMLObjectTypes.

    Note:
        `ids` and `object_types` must have the same `len` to be zipped together.

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        ids (list): List of Object IDs
        object_types (list): List of Object Types
        project_id (str): ID of a project in which the objects are located.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    if (
        not ids
        or not object_types
        or not (zipped := zip(ids, object_types, strict=True))
    ):
        raise ValueError("Both ids and object_types must be provided.")

    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id

    body = [{"did": did, "tp": tp, "pip": project_id} for did, tp in zipped]
    body = {"delete": body}
    return connection.post(
        endpoint='/api/objects/deleteObjects',
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
    )


@ErrorHandler(err_msg="Error updating object with ID {id}")
def update_object(
    connection: Connection,
    id,
    body,
    object_type,
    project_id=None,
    error_msg=None,
    verbose=True,
):
    """Get information for a specific object in a specific project; if you do
    not specify a project ID, you get information for the object in all
    projects.

    You identify the object with the object ID and object type. You specify
    the object type as a query parameter; possible values for object type are
    provided in EnumDSSXMLObjectTypes.

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        body: (object): body of the response
        object_type (int): One of EnumDSSXMLObjectTypes. Ex. 34 (User or
        UserGroup), 44 (Security Role), 32 (Project), 8 (Folder), 36 (type of
        I-Server configuration)
        project_id(str): ID of a project in which the object is located.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    project_id = _validate_project_id(connection, object_type, project_id, False)

    return connection.put(
        endpoint=f'/api/objects/{id}',
        headers={'X-MSTR-ProjectID': project_id},
        params={'type': object_type},
        json=body,
    )


@ErrorHandler(err_msg="Error creating a copy of object with ID {id}")
def copy_object(
    connection: Connection,
    id,
    name,
    folder_id,
    object_type,
    project_id=None,
    error_msg=None,
):
    """Create a copy of a specific object.

    You identify the object with the object ID and object type. You obtain the
    authorization token needed to execute the request using POST /auth/login;
    you obtain the project ID using GET /projects. You pass the authorization
    token and the project ID in the request header. You specify the object ID in
    the path of the request and object type as a query parameter; possible
    values for object type are provided in EnumDSSXMLObjectTypes. You specify
    the name and location (folder ID) of the new object in the body of the
    request. If you do not specify a new name, a default name is generated, such
    as 'Old Name (1)'. If you do not specify a folder ID, the object is saved in
    the same folder as the source object.

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        object_type (int): One of EnumDSSXMLObjectTypes. Ex. 34 (User or
        UserGroup), 44 (Security Role), 32 (Project), 8 (Folder), 36 (type of
        I-Server configuration)
        project_id(str): ID of a project in which the object is located.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    project_id = _validate_project_id(connection, object_type, project_id, False)

    if not project_id:
        raise ValueError("Project needs to be specified.")

    body = {'name': name, 'folderId': folder_id}
    return connection.post(
        endpoint=f'/api/objects/{id}/copy',
        headers={'X-MSTR-ProjectID': project_id},
        params={'type': object_type},
        json=body,
    )


@ErrorHandler(err_msg="Error getting property set for object with ID {id}")
def get_property_set(
    connection: Connection,
    id: str,
    obj_type: int,
    property_set_id: str,
    error_msg: str | None = None,
):
    """Update a property set for an object.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project id string
        obj_type (int): Object type, as in ObjectTypes enum
        property_set_id (string): ID of a property set
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f"/api/objects/{id}/type/{obj_type}"
        f"/propertySets/{property_set_id}/properties",
    )


@ErrorHandler(err_msg="Error updating property set for object with ID {id}")
def update_property_set(
    connection: Connection,
    id: str,
    obj_type: int,
    body: dict,
    error_msg: str | None = None,
):
    """Update a property set for an object.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (string): Project id string
        obj_type (int): Object type, as in ObjectTypes enum
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.put(
        endpoint=f"/api/objects/{id}/type/{obj_type}/propertySets",
        json=body,
    )


@ErrorHandler(err_msg="Error getting VLDB settings for object with ID {id}")
def get_vldb_settings(
    connection: Connection, id, object_type, project_id=None, error_msg=None
):
    """Get vldb settings for an object.

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        object_type (int): DssXmlTypeReportDefinition(3) for Dataset and
            DssXmlTypeDocumentDefinition(55) for dashboard/document
        project_id(str): ID of a project in which the object is located.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server
    """
    headers = {}
    if project_id:
        headers = {'X-MSTR-ProjectID': project_id}
    else:
        connection._validate_project_selected()
        headers = {'X-MSTR-ProjectID': connection.project_id}

    return connection.get(
        endpoint=f'/api/objects/{id}/vldb/propertySets',
        params={'type': object_type},
        headers=headers,
    )


@ErrorHandler(
    err_msg="Error resetting all custom vldb settings for object with ID {id}"
)
def delete_vldb_settings(
    connection: Connection, id, object_type, project_id=None, error_msg=None
):
    """Delete all customized vldb settings in one object, this operation will
    reset all vldb settings to default.

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        object_type (int): DssXmlTypeReportDefinition(3) for Dataset and
            DssXmlTypeDocumentDefinition(55) for dashboard/document
        project_id(str): ID of a project in which the object is located.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server
    """
    headers = {}
    if project_id:
        headers = {'X-MSTR-ProjectID': project_id}
    else:
        connection._validate_project_selected()
        headers = {'X-MSTR-ProjectID': connection.project_id}

    return connection.delete(
        endpoint=f'/api/objects/{id}/vldb/propertySets',
        params={'type': object_type},
        headers=headers,
    )


@ErrorHandler(
    err_msg="Error resetting all custom vldb settings for object with ID {id}"
)
def set_vldb_settings(
    connection: Connection, id, object_type, name, body, project_id=None, error_msg=None
):
    """Set vldb settings for one property set in one object.

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        object_type (int): DssXmlTypeReportDefinition(3) for Dataset and
            DssXmlTypeDocumentDefinition(55) for dashboard/document
        name: property set name
        body: [{"name": "string",
                "value": {}}]
        project_id(str): ID of a project in which the object is located.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server
    """
    headers = {}
    if project_id:
        headers = {'X-MSTR-ProjectID': project_id}
    else:
        connection._validate_project_selected()
        headers = {'X-MSTR-ProjectID': connection.project_id}

    return connection.put(
        endpoint=f'/api/objects/{id}/vldb/propertySets/{name}',
        params={'type': object_type},
        headers=headers,
        json=body,
    )


@ErrorHandler(err_msg="Error getting objects.")
def create_search_objects_instance(
    connection: Connection,
    name=None,
    pattern=4,
    domain=2,
    root=None,
    object_type=None,
    error_msg=None,
):
    """Create a search instance.

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        name: expression used with the pattern to do the search
        pattern: specifies the nature of the search. Possible values are defined
            in the EnumDSSXMLSearchTypes javadoc
        domain: search domain. specifies the domain/scope of the search.
            Possible values are defined in the EnumDSSXMLSearchDomain javadoc
        root: folder ID of the root in which the search is done
        object_type: specifies the type of objects to be searched. Possible
            values are defined in the EnumDSSObjectType javadoc
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response returned by the Strategy One REST server
    """
    connection._validate_project_selected()
    return connection.post(
        endpoint='/api/objects',
        headers={'X-MSTR-ProjectID': connection.project_id},
        params={
            'name': name,
            'pattern': pattern,
            'domain': domain,
            'root': root,
            'type': object_type,
        },
    )


@ErrorHandler(err_msg="Error getting objects using search with ID {search_id}")
def get_objects(
    connection: Connection,
    search_id,
    offset=0,
    limit=-1,
    get_tree=False,
    error_msg=None,
):
    """Get list of objects from metadata.

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        search_id: ID for the results of a previous search stored in I-Server
            memory
        offset: starting point within the collection of returned results. Used
            to control paging behavior.
        limit: maximum number of items returned for a single request. Used to
            control paging behavior
        get_tree: specifies that the search results should be displayed in
            a tree structure instead of a list. The ancestors of the searched
            objects are the nodes and the searched objects are the leaves of
            the tree.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response returned by the Strategy One REST server
    """
    connection._validate_project_selected
    return connection.get(
        endpoint='/api/objects',
        headers={'X-MSTR-ProjectID': connection.project_id},
        params={
            'searchId': search_id,
            'offset': offset,
            'limit': limit,
            'getTree': get_tree,
        },
    )


def get_objects_async(
    future_session: 'FuturesSessionWithRenewal',
    search_id,
    offset=0,
    limit=-1,
    get_tree=False,
):
    """Get list of objects from metadata asynchronously.

    Args:
        future_session(object): Future Session object to call Strategy One REST
            Server asynchronously
        search_id: ID for the results of a previous search stored in I-Server
            memory
        offset: starting point within the collection of returned results. Used
            to control paging behavior.
        limit: maximum number of items returned for a single request. Used to
            control paging behavior.
        get_tree: specifies that the search results should be displayed in
            a tree structure instead of a list. The ancestors of the searched
            objects are the nodes and the searched objects are the leaves of
            the tree.

    Returns:
        HTTP response returned by the Strategy One REST server
    """
    future_session.connection._validate_project_selected()
    endpoint = '/api/objects'
    headers = {'X-MSTR-ProjectID': future_session.connection.project_id}
    params = {
        'searchId': search_id,
        'offset': offset,
        'limit': limit,
        'getTree': get_tree,
    }
    future = future_session.get(endpoint=endpoint, headers=headers, params=params)
    return future


@ErrorHandler(err_msg="Error certifying object with ID {id}")
def toggle_certification(connection: Connection, id, object_type=3, certify=True):
    """Certify/Uncertify a multi-table dataset.

    Args:
        connection (object): Strategy One connection object returned by
            `connection.Connection()`.
        id (str): Identifier of a pre-existing dataset. Used when
            certifying a pre-existing dataset.
        object_type (int, optional): Type of object to certify as integer;
            defaults to 3 (dataset)
        certify (bool, optional): boolean representing if the instruction is to
            certify (True) or decertify (False); defaults to True.

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    endpoint = (
        f'/api/objects/{id}/certify/?type={str(object_type)}&certify={str(certify)}'
    )
    return connection.put(
        endpoint=endpoint,
        headers={'Content-Type': 'application/json', 'Accept': 'application/json'},
    )


@ErrorHandler(err_msg='Error updating translations for object with ID {id}')
def update_translations(
    connection: Connection,
    project_id: str,
    id: str,
    object_type: int,
    body: dict,
    fields,
):
    """Update translations for a specific object.
    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        project_id (str): ID of the project in which the object is located
        id (str): ID of the object
        object_type (int): Type of the object
        body (json): body of the request
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model
    Returns:
        HTTP response object returned by the Strategy One REST server."""

    object_type = ObjectTypes(object_type).name
    return connection.patch(
        endpoint=f'/api/objects/{object_type}/{id}/translations',
        headers={'X-MSTR-ProjectID': project_id},
        params={'fields': fields},
        json=body,
    )


@ErrorHandler(err_msg='Error getting translations for object with ID {id}')
def get_translations(
    connection: Connection, project_id: str | None, id: str, object_type: int, fields
):
    """Get translations for a specific object.
    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        project_id (str | None): ID of the project in which the object is
            located in.
        id (str): ID of the object
        object_type (int): Type of the object
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model
    Returns:
        HTTP response object returned by the Strategy One REST server."""

    object_type = ObjectTypes(object_type).name
    return connection.get(
        endpoint=f'/api/objects/{object_type}/{id}/translations',
        headers={'X-MSTR-ProjectID': project_id},
        params={'fields': fields},
    )


@ErrorHandler(err_msg='Error creating shortcut for object with ID {id}')
def create_shortcut(
    connection: Connection,
    project_id: str,
    id: str,
    object_type: int,
    body: dict,
):
    """Create a shortcut to an object.

    Args:
        connection: Strategy One REST API connection object
        id: ID of the object to create a shortcut to
        object_type: Type of the object to create a shortcut to
        body: A dictionary specifying the request body

    Returns:
        Complete HTTP response object.
    """
    return connection.post(
        endpoint=f'/api/objects/{id}/type/{object_type}/shortcuts',
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
    )



from mstrio.api import browsing as browsing_api
from mstrio.connection import Connection


def get_objects_from_quick_search(
    connection: 'Connection',
    body: dict,
    include_ancestors: bool = False,
    show_navigation_path: bool = False,
    fields: str | None = None,
):
    return (
        browsing_api.get_objects_from_quick_search(
            connection=connection,
            body=body,
            include_ancestors=include_ancestors,
            show_navigation_path=show_navigation_path,
            fields=fields,
        )
        .json()
        .get('result')
    )


def get_search_object(
    connection: 'Connection',
    id: str,
):
    result_dict: dict = browsing_api.get_search_object(
        connection=connection, id=id
    ).json()

    time_range = result_dict.pop('timeRange', None)
    date_filter_type = result_dict.pop('dateFilterType', None)
    if date_filter_type and date_filter_type.upper() == 'CREATED':
        result_dict['dateCreatedQuery'] = time_range
    elif date_filter_type and date_filter_type.upper() == 'MODIFIED':
        result_dict['dateModifiedQuery'] = time_range

    search_visibility = result_dict.pop('visibility', None)
    if search_visibility and search_visibility.upper() == 'VISIBLE':
        result_dict['includeHidden'] = False
    elif search_visibility and search_visibility.upper() == 'ALL':
        result_dict['includeHidden'] = True

    if 'ownerId' in result_dict:
        result_dict['ownerQuery'] = result_dict.pop('ownerId')

    return result_dict



"""Create a bursting subscription.
1. Connect to the environment using data from workstation
2. Create a bursting configuration
   - 2.1 using objects
   - 2.2 using IDs
3. Create an email subscription with created bursting configuration
"""

from mstrio.connection import get_connection
from mstrio.distribution_services import Content, Device, EmailSubscription, Schedule
from mstrio.modeling import Attribute, AttributeForm

conn = get_connection(workstationData, project_name='MicroStrategy Tutorial')


# Create bursting configuration, using objects
bursting = Content.Properties.Bursting(
    slicing_attributes=[Attribute(conn, id='8D679D4311D3E4981000E787EC6DE8A4')],
    address_attribute=Attribute(conn, id='8D679D4311D3E4981000E787EC6DE8A4'),
    device=Device(conn, id='1D2E6D168A7711D4BE8100B0D04B6F0B'),
    form=AttributeForm(conn, id='8A9EFEDB11D60C62C000CC8F9590594F'),
)

# Alternatively, create bursting configuration using IDs.
# However, the `slicing_attributes` should be provided in the format:
# '<attribute_id>~<attribute_name>'
# bursting = Content.Properties.Bursting(
#     slicing_attributes=['8D679D4311D3E4981000E787EC6DE8A4~Manager'],
#     address_attribute_id='8D679D4311D3E4981000E787EC6DE8A4',
#     device_id='1D2E6D168A7711D4BE8100B0D04B6F0B',
#     form_id='8A9EFEDB11D60C62C000CC8F9590594F',
# )

# Create bursting subscription
bursting_subscription = EmailSubscription.create(
    connection=conn,
    name='Bursting Subscription',
    project_name='MicroStrategy Tutorial',
    contents=[
        Content(
            id='<content_id>',  # Content should have bursting enabled
            type=Content.Type.REPORT,
            personalization=Content.Properties(
                format_type=Content.Properties.FormatType.PDF,
                bursting=bursting,
            ),
        ),
    ],
    schedules=Schedule(conn, name='Monday Morning'),
    recipients=['54F3D26011D2896560009A8E67019608'],  # Administrator
    email_subject='Bursting Subscription',
)



from dataclasses import dataclass

from mstrio.utils.helper import Dictable


@dataclass
class Color(Dictable):
    """A class containing information about the color in different formats.
        It is used as a value for `FormatProperties` that type related
        with colors.

    Attributes:
        hex_value(str): color expressed in hex format (e.g. '#ff02ef')
        server_value(str): color translated for server readable format
        red(str): red component of a RGB format, expressed as decimal
        green(str): green component of a RGB format, expressed as decimal
        blue(str): blue component of a RGB format, expressed as decimal"""

    def _rgb_base_converter(
        self, r: str, g: str, b: str, original_base: int, desired_base: int
    ):
        format_spec = {2: '0>8b', 10: '', 16: '0>2x'}
        return (
            format(int(r, original_base), format_spec[desired_base]),
            format(int(g, original_base), format_spec[desired_base]),
            format(int(b, original_base), format_spec[desired_base]),
        )

    def _init_from_hex(self, hex_val: str) -> str:
        r, g, b = self._rgb_base_converter(
            r=hex_val[1:3],
            g=hex_val[3:5],
            b=hex_val[5:7],
            original_base=16,
            desired_base=2,
        )
        self.server_value = str(int(r + g + b, 2))
        self.hex_value = hex_val
        self.red, self.green, self.blue = self._rgb_base_converter(r, g, b, 2, 10)

    def _init_from_rest(self, rest_val: str) -> str:
        bin_color = format(int(rest_val), '0>24b')
        r, g, b = self._rgb_base_converter(
            r=bin_color[0:8],
            g=bin_color[8:16],
            b=bin_color[16:24],
            original_base=2,
            desired_base=16,
        )
        self.server_value = rest_val
        self.hex_value = f'#{r}{g}{b}'
        self.red, self.green, self.blue = self._rgb_base_converter(r, g, b, 16, 10)

    def _init_from_rgb(self, red: str, green: str, blue: str) -> str:
        r, g, b = self._rgb_base_converter(
            r=red, g=green, b=blue, original_base=10, desired_base=2
        )
        self.server_value = str(int(r + g + b, 2))
        r, g, b = self._rgb_base_converter(
            r=red, g=green, b=blue, original_base=10, desired_base=16
        )
        self.hex_value = f'#{r}{g}{b}'
        self.red, self.green, self.blue = red, green, blue

    hex_value: str
    server_value: str
    red: str
    green: str
    blue: str

    def __init__(
        self,
        hex_value: str | None = None,
        red: int | None = None,
        green: int | None = None,
        blue: int | None = None,
        server_value: str | None = None,
    ):
        """Create an object representing color value of a `FormatProperty`.
            It can be created by providing either hex value, server readable
            value or all three RGB components. Two other representation will
            be generated automatically based on the one provided by the user.
            If more than one representation provided, priority goes as follows:
            server_value->hex_value->RGB values.

        Args:
            hex_value(str): color expressed in hex format (e.g. '#ff02ef')
            server_value(str): color translated for server readable format
            red(int): component of a RGB format, expressed as decimal
            green(int): component of a RGB format, expressed as decimal
            blue(int): component of a RGB format, expressed as decimal"""

        if (
            server_value is not None
            and isinstance(server_value, str)
            and server_value.isnumeric()
        ):
            self._init_from_rest(rest_val=server_value)
        elif hex_value is not None and hex_value[0] == '#' and len(hex_value) == 7:
            self._init_from_hex(hex_val=hex_value)
        elif red in range(256) and green in range(256) and blue in range(256):
            self._init_from_rgb(str(red), str(green), str(blue))
        else:
            raise ValueError(
                'Invalid parameter for Color value of the format property.'
            )

    def __repr__(self):
        return (
            f'hex_value: {self.hex_value}, rgb: ({self.red}, {self.green}, '
            f'{self.blue}), server_value: {self.server_value}'
        )



from enum import auto

from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import Dictable, exception_handler


class EmailFormat(AutoName):
    UU_ENCODED = auto()
    MIME = auto()
    UNSUPPORTED = auto()


class EmailEncoding(AutoName):
    QUOTED_PRINTABLE = auto()
    BASE64 = auto()
    ASCII_7BIT = auto()
    UNSUPPORTED = auto()


class EmailMessageSensitivity(AutoName):
    NORMAL = auto()
    PERSONAL = auto()
    PRIVATE = auto()
    CONFIDENTIAL = auto()
    UNSUPPORTED = auto()


class FtpProtocol(AutoName):
    FTP = auto()
    FTP_OVER_SSL_TLS_EXPLICIT = auto()
    FTP_OVER_SSL_TLS_IMPLICIT = auto()
    SFTP = auto()
    UNSUPPORTED = auto()


class FileEncoding(AutoName):
    ENCRYPT = auto()
    COMPRESS = auto()
    NO_ENCRYPT_COMPRESS = auto()


class PrinterPaperSource(AutoName):
    AUTOMATIC = auto()
    MANUAL = auto()
    UNSUPPORTED = auto()


class PrinterPaperSize(AutoName):
    LETTER = auto()
    LETTER_SMALL = auto()
    TABLOID = auto()
    LEDGER = auto()
    LEGAL = auto()
    STATEMENT = auto()
    EXECUTIVE = auto()
    A3 = auto()
    A4 = auto()
    A4_SMALL = auto()
    A5 = auto()
    B4 = auto()
    B5 = auto()
    FOLIO = auto()
    UNSUPPORTED = auto()


class PrinterBackupLocationType(AutoName):
    PRINTER = auto()
    FILE = auto()


class PdfOddEvenPages(AutoName):
    ODD_PAGES = auto()
    EVEN_PAGES = auto()
    ALL_PAGES = auto()
    UNSUPPORTED = auto()


class PdfApplicationPriority(AutoName):
    OTHER = auto()
    I_SERVER = auto()
    UNSUPPORTED = auto()


class FileSystem(Dictable):
    """File System Options

    Attributes:
        create_folder: Whether to create required folder in the path,
            default: True
        filename_append_time_stamp: Whether to append timestamp to the file
            name, default: True
        override_filename: Whether to override filename with same name,
            default: False
        append_to_file: Whether to append to the file, default: False
    """

    def __init__(
        self,
        create_folder: bool = True,
        filename_append_time_stamp: bool = True,
        override_filename: bool = False,
        append_to_file: bool = False,
    ):
        self.create_folder = create_folder
        self.filename_append_time_stamp = filename_append_time_stamp
        self.override_filename = override_filename
        self.append_to_file = append_to_file


class ConnectionParameters(Dictable):
    """Connection Parameters

    Attributes:
        retries_count: Number of Retries, default: 5
        seconds_between_retries: Time between retries (Seconds), default: 3
        delivery_timeout_seconds: Delivery Timeout (Seconds), default: 10
    """

    def __init__(
        self,
        retries_count: int = 5,
        seconds_between_retries: int = 3,
        delivery_timeout_seconds: int = 10,
    ):
        self.retries_count = retries_count
        self.seconds_between_retries = seconds_between_retries
        self.delivery_timeout_seconds = delivery_timeout_seconds


class EmailMimeSettings(Dictable):
    """MIME settings for MIME email format

    Attributes:
        plain_text_html_body_encoding: Indicates the encoding for plaintext and
            HTML body, EmailEncoding enum
        text_attachment_encoding: Indicates the encoding for text attachment,
            only quoted_printable and base64 are supported, EmailEncoding enum
        binary_attachment_encoding: Indicates the encoding for text attachment,
            only quoted_printable and base64 are supported, EmailEncoding enum
        message_sensitivity: Indicates the message sensitivity,
            EmailMessageSensitivity enum
        us_ascii_encoding: Whether to use US-ASCII for Subject,
            Attachments Names and Display Names, default: False
        non_us_ascii_quotes: Whether to use put quotes around non US-ASCII
            Display Names. Enable for Microsoft Outlook 98 and
            Microsoft Outlook 2000 email client only, default: False
        embed_html_attachments: Whether to embed HTML Attachments,
            default: False
        embed_all_attachments: Whether to embed All Attachments, default: False
        embed_adobe_flash_content: Whether to embed
            Adobe Flash Content, default: False
        html_table_position_only: Whether to use only table to position
            elements in HTML (Enabled for Microsoft Outlook 2007),
            default: False
        css_inline_style: Whether to use inline style CSS, default: False
    """

    _FROM_DICT_MAP = {
        "plain_text_html_body_encoding": EmailEncoding,
        "text_attachment_encoding": EmailEncoding,
        "binary_attachment_encoding": EmailEncoding,
        "message_sensitivity": EmailMessageSensitivity,
    }

    def __init__(
        self,
        plain_text_html_body_encoding: EmailEncoding | str,
        text_attachment_encoding: EmailEncoding | str,
        binary_attachment_encoding: EmailEncoding | str,
        message_sensitivity: EmailMessageSensitivity | str,
        us_ascii_encoding: bool = False,
        non_us_ascii_quotes: bool = False,
        embed_html_attachments: bool = False,
        embed_all_attachments: bool = False,
        embed_adobe_flash_content: bool = False,
        html_table_position_only: bool = False,
        css_inline_style: bool = False,
    ):
        self.plain_text_html_body_encoding = (
            EmailEncoding(plain_text_html_body_encoding)
            if isinstance(plain_text_html_body_encoding, str)
            else plain_text_html_body_encoding
        )
        self.text_attachment_encoding = (
            EmailEncoding(text_attachment_encoding)
            if isinstance(text_attachment_encoding, str)
            else text_attachment_encoding
        )
        self.binary_attachment_encoding = (
            EmailEncoding(binary_attachment_encoding)
            if isinstance(binary_attachment_encoding, str)
            else binary_attachment_encoding
        )
        self.message_sensitivity = (
            EmailMessageSensitivity(message_sensitivity)
            if isinstance(message_sensitivity, str)
            else message_sensitivity
        )
        self.us_ascii_encoding = us_ascii_encoding
        self.non_us_ascii_quotes = non_us_ascii_quotes
        self.embed_html_attachments = embed_html_attachments
        self.embed_all_attachments = embed_all_attachments
        self.embed_adobe_flash_content = embed_adobe_flash_content
        self.html_table_position_only = html_table_position_only
        self.css_inline_style = css_inline_style


class EmailSmartHostSettings(Dictable):
    """Smart host settings

    Attributes:
        server: Server Name or IP Address
        port: Port number, default: 25
        always_use_smart_host: Whether to always use smart host, default: 25
        smart_host_username: Smart Host Username
        smart_host_password: Smart Host Password
    """

    def __init__(
        self,
        server: str | None = None,
        port: int | None = None,
        always_use_smart_host: bool = False,
        smart_host_username: str | None = None,
        smart_host_password: bool = False,
    ):
        self.server = server
        self.port = port
        self.always_use_smart_host = always_use_smart_host
        self.smart_host_username = smart_host_username
        self.smart_host_password = smart_host_password


class EmailDeviceProperties(Dictable):
    """Device properties for email device type.

    Attributes:
        format: indicates the email format of the email device, EmailFormat enum
        mime_settings: MIME settings for MIME email format
        smart_host_settings: smart host settings
    """

    _FROM_DICT_MAP = {
        "format": EmailFormat,
        "mime_settings": EmailMimeSettings.from_dict,
        "smart_host_settings": EmailSmartHostSettings.from_dict,
    }

    def __init__(
        self,
        format: str | EmailFormat,
        mime_settings: dict | EmailMimeSettings | None = None,
        smart_host_settings: dict | EmailSmartHostSettings | None = None,
    ):
        self.format = EmailFormat(format) if isinstance(format, str) else format
        self.mime_settings = (
            EmailMimeSettings.from_dict(mime_settings)
            if isinstance(mime_settings, dict)
            else mime_settings
        )
        self.smart_host_settings = (
            EmailSmartHostSettings.from_dict(smart_host_settings)
            if isinstance(smart_host_settings, dict)
            else smart_host_settings
        )


class FtpServerSettings(Dictable):
    """FTP Server Settings

    Attributes:
        protocol: FTP Protocol, FtpProtocol enum
        host: FTP Host IP Address
        port: Port number
        path: FTP Path
        passive_mode: Whether to use Passive Mode for connection, default: True
        max_connections: Maximum connection, default: -1
        ascii_mode: Whether to use ASCII Mode File Type for connection,
            default: True
        user_name: FTP Account Username
        password: FTP Account Password
    """

    _FROM_DICT_MAP = {
        "protocol": FtpProtocol,
    }

    def __init__(
        self,
        protocol: FtpProtocol | str,
        port: int,
        path: str,
        host: str | None = None,
        passive_mode: bool = True,
        max_connections: int = -1,
        ascii_mode: bool = True,
        user_name: str | None = None,
        password: str | None = None,
    ):
        self.protocol = FtpProtocol(protocol) if isinstance(protocol, str) else protocol
        self.host = host
        self.port = port
        self.path = path
        self.passive_mode = passive_mode
        self.max_connections = max_connections
        self.ascii_mode = ascii_mode
        self.user_name = user_name
        self.password = password


class FtpDeviceProperties(Dictable):
    """Device properties for ftp device type.

    Attributes:
        server_settings: FTP Server Settings, FtpServerSettings class
        file_system: File System Options, FileSystem class
    """

    _FROM_DICT_MAP = {
        "server_settings": FtpServerSettings.from_dict,
        "file_system": FileSystem.from_dict,
    }

    def __init__(
        self,
        server_settings: FtpServerSettings | dict,
        file_system: FileSystem | dict,
    ):
        self.server_settings = (
            FtpServerSettings.from_dict(server_settings)
            if isinstance(server_settings, dict)
            else server_settings
        )
        self.file_system = (
            FileSystem.from_dict(file_system)
            if isinstance(file_system, dict)
            else file_system
        )


class IOSDeviceProperties(Dictable):
    """Device properties for ipad and iphone device types.

    Attributes:
        app_id: Application ID
        server: Push notification server name, default: 'gateway.push.apple.com'
        port: Port number, default: 2195
        provider_certificate: Provider Certificate file location on server
        feedback_service_server: Feedback Service server name,
            default: 'feedback.push.apple.com'
        feedback_service_port: Feedback Service Port number, default: 2195
    """

    def __init__(
        self,
        app_id: str = '',
        server: str | None = None,
        port: int | None = None,
        provider_certificate: str | None = None,
        feedback_service_server: str | None = None,
        feedback_service_port: int | None = None,
    ):
        self.app_id = app_id
        self.server = server
        self.port = port
        self.provider_certificate = provider_certificate
        self.feedback_service_server = feedback_service_server
        self.feedback_service_port = feedback_service_port


class AndroidDeviceProperties(Dictable):
    """Device properties for android device type.

    Attributes:
        package_name: Mobile Application Package name
        server: Firebase cloud messaging service server name/IP address
        port: Firebase cloud messaging service Port number, default: 0
        auth_key: Firebase cloud messaging service Authentication Key
        collapse_key: Firebase cloud messaging service Collapse Key
        delay_with_idle: Whether to use Delay with idle
        notification_active_hours: Notification active for hours, 1 to 99999
        use_proxy: Whether to use Proxy for Firebase cloud messaging
        proxy_server: Firebase cloud messaging proxy service server
            name/IP address
        proxy_port: Firebase cloud messaging proxy service Port number
    """

    def __init__(
        self,
        package_name: str,
        server: str | None = None,
        port: int | None = None,
        auth_key: str | None = None,
        collapse_key: str | None = None,
        delay_with_idle: bool = False,
        notification_active_hours: int | None = None,
        use_proxy: bool = False,
        proxy_server: str | None = None,
        proxy_port: int | None = None,
    ):
        self.package_name = package_name
        self.server = server
        self.port = port
        self.auth_key = auth_key
        self.collapse_key = collapse_key
        self.delay_with_idle = delay_with_idle
        self.notification_active_hours = notification_active_hours
        self.use_proxy = use_proxy
        self.proxy_server = proxy_server
        self.proxy_port = proxy_port


class FileLocation(Dictable):
    """File Location Setting

    Attributes:
        file_path: File Location
        append_user_path: Whether to allow and append user entered path,
            default: False
        use_backup_location: Whether to use backup location, default: False
        backup_file_path: Backup File Location
    """

    def __init__(
        self,
        file_path: str,
        append_user_path: bool = False,
        use_backup_location: bool = False,
        backup_file_path: str | None = None,
    ):
        self.file_path = file_path
        self.append_user_path = append_user_path
        self.use_backup_location = use_backup_location
        self.backup_file_path = backup_file_path


class FileProperties(Dictable):
    """File related properties

    Attributes:
        read_only: Whether to enable read only access for file, default: True
        archive: Whether to archive the file, default: False
        index: Whether to index the file, default: False
        file_encoding: File Encoding, FileEncoding enum
        unix_access_rights: UNIX access rights
    """

    _FROM_DICT_MAP = {
        "file_encoding": FileEncoding,
    }

    def __init__(
        self,
        read_only: bool = True,
        archive: bool = False,
        index: bool = False,
        file_encoding: FileEncoding | str | None = None,
        unix_access_rights: str | None = None,
    ):
        self.read_only = read_only
        self.archive = archive
        self.index = index
        self.file_encoding = (
            FileEncoding(file_encoding)
            if isinstance(file_encoding, str)
            else file_encoding
        )
        self.unix_access_rights = unix_access_rights


class UnixWindowsSharity(Dictable):
    """Unix to Windows Sharity settings

    Attributes:
        sharity_enabled: Whether to enable delivery from Intelligence
            Server running on UNIX to Windows, default: False
        server_username: IServer Username used for Sharity
        server_password: IServer Password used for Sharity
        server_mount_root: IServer Mount root folder path for Sharity
    """

    def __init__(
        self,
        sharity_enabled: bool = False,
        server_username: str | None = None,
        server_password: str | None = None,
        server_mount_root: str | None = None,
    ):
        self.sharity_enabled = sharity_enabled
        self.server_username = server_username
        self.server_password = server_password
        self.server_mount_root = server_mount_root


class FileDeviceProperties(Dictable):
    """Device properties for file device type.

    Attributes:
        file_location: File Location Setting, FileLocation class
        file_system: File System Options, FileSystem class
        connection_parameters: Connection Parameters,
            ConnectionParameters class
        file_properties: File related properties, FileProperties class
        unix_windows_sharity: Unix to Windows Sharity settings,
            UnixWindowsSharity class
    """

    _FROM_DICT_MAP = {
        "file_location": FileLocation.from_dict,
        "file_system": FileSystem.from_dict,
        "connection_parameters": ConnectionParameters.from_dict,
        "file_properties": FileProperties.from_dict,
        "unix_windows_sharity": UnixWindowsSharity.from_dict,
    }

    def __init__(
        self,
        file_location: FileLocation | dict,
        file_system: FileSystem | dict,
        connection_parameters: ConnectionParameters | dict,
        file_properties: FileProperties | dict,
        unix_windows_sharity: UnixWindowsSharity | dict,
    ):
        self.file_location = (
            FileLocation.from_dict(file_location)
            if isinstance(file_location, dict)
            else file_location
        )
        self.file_system = (
            FileSystem.from_dict(file_system)
            if isinstance(file_system, dict)
            else file_system
        )
        self.connection_parameters = (
            ConnectionParameters.from_dict(connection_parameters)
            if isinstance(connection_parameters, dict)
            else connection_parameters
        )
        self.file_properties = (
            FileProperties.from_dict(file_properties)
            if isinstance(file_properties, dict)
            else file_properties
        )
        self.unix_windows_sharity = (
            UnixWindowsSharity.from_dict(unix_windows_sharity)
            if isinstance(unix_windows_sharity, dict)
            else unix_windows_sharity
        )


class PrinterLocation(Dictable):
    """Printer Location.

    Attributes:
        location: Printer Location
        user_defined_location: Whether to allow user defined location,
            default: False
    """

    def __init__(
        self, location: str | None = None, user_defined_location: bool = False
    ):
        self.location = location
        self.user_defined_location = user_defined_location


class PrinterPdfSettings(Dictable):
    """PDF printing settings.

    Attributes:
        post_script_level: Post Script Level, 0 to 9
        odd_even_pages: Indicates the odd or even printing of pages,
            PdfOddEvenPages enum
        reverse_pages: Whether to use reverse pages, default: False
        application_priority: Indicates the which application should get the
            priority when another application is using Adobe and IServer is
            trying to print PDF documents, PdfApplicationPriority enum
    """

    _FROM_DICT_MAP = {
        "odd_even_pages": PdfOddEvenPages,
        "application_priority": PdfApplicationPriority,
    }

    def __init__(
        self,
        post_script_level: int = 2,
        reverse_pages: bool = False,
        odd_even_pages: PdfOddEvenPages | str | None = None,
        application_priority: PdfApplicationPriority | str | None = None,
    ):
        self.post_script_level = (
            post_script_level
            if 0 <= post_script_level <= 9
            else exception_handler(
                'Please provide appropriate value for post_script_level (from 0 to 9)',
                ValueError,
            )
        )
        self.odd_even_pages = (
            PdfOddEvenPages(odd_even_pages)
            if isinstance(odd_even_pages, str)
            else odd_even_pages
        )
        self.reverse_pages = reverse_pages
        self.application_priority = (
            PdfApplicationPriority(application_priority)
            if isinstance(application_priority, str)
            else application_priority
        )


class PrinterProperties(Dictable):
    """Printer Properties.

    Attributes:
        pdf_setting: PDF printing settings, PrinterPdfSettings class
        quality: Print Quality
        scale: Scale in %, default: 100
        paper_source: Paper Source, PrinterPaperSource class
        paper_size: Paper Size, PrinterPaperSize class
    """

    _FROM_DICT_MAP = {
        "pdf_setting": PrinterPdfSettings.from_dict,
        "paper_source": PrinterPaperSource,
        "paper_size": PrinterPaperSize,
    }

    def __init__(
        self,
        pdf_setting: PrinterPdfSettings | dict,
        quality: str | None = None,
        scale: int = 100,
        paper_source: PrinterPaperSource | str | None = None,
        paper_size: PrinterPaperSize | str | None = None,
    ):
        self.pdf_setting = (
            PrinterPdfSettings.from_dict(pdf_setting)
            if isinstance(pdf_setting, dict)
            else pdf_setting
        )
        self.quality = quality
        self.scale = scale
        self.paper_source = (
            PrinterPaperSource(paper_source)
            if isinstance(paper_source, str)
            else paper_source
        )
        self.paper_size = (
            PrinterPaperSize(paper_size) if isinstance(paper_size, str) else paper_size
        )


class BackupPrinterProperties(Dictable):
    """Printer Backup Setting.

    Attributes:
        print_on_backup: Whether to print on Backup printer if
            primary printer fails, default: False
        backup_location_type: Backup device type,
            PrinterBackupLocationType class
        backup_printer_location: Backup Printer Location
        backup_file_location: Backup File Location
    """

    _FROM_DICT_MAP = {
        "backup_location_type": PrinterBackupLocationType,
    }

    def __init__(
        self,
        print_on_backup: bool = False,
        backup_location_type: PrinterBackupLocationType | str | None = None,
        backup_printer_location: str | None = None,
        backup_file_location: str | None = None,
    ):
        self.print_on_backup = print_on_backup
        self.backup_location_type = (
            PrinterBackupLocationType(backup_location_type)
            if isinstance(backup_file_location, str)
            else backup_location_type
        )
        self.backup_printer_location = backup_printer_location
        self.backup_file_location = backup_file_location


class PrinterDeviceProperties(Dictable):
    """Device properties for printer device type.

    Attributes:
        printer_location: Printer Location, PrinterLocation class
        printer_properties: Printer Properties, PrinterProperties class
        connection_parameters: Connection Parameters, ConnectionParameters class
        backup_printer_properties: Printer Backup Setting,
            BackupPrinterProperties class
        temp_file_location: Temp File Location during print execution
    """

    _FROM_DICT_MAP = {
        "printer_location": PrinterLocation.from_dict,
        "printer_properties": PrinterProperties.from_dict,
        "connection_parameters": ConnectionParameters.from_dict,
        "backup_printer_properties": BackupPrinterProperties.from_dict,
    }

    def __init__(
        self,
        printer_location: PrinterLocation | dict,
        printer_properties: PrinterProperties | dict,
        connection_parameters: ConnectionParameters | dict,
        backup_printer_properties: BackupPrinterProperties | dict,
        temp_file_location: str | None = None,
    ):
        self.printer_location = (
            PrinterLocation.from_dict(printer_location)
            if isinstance(printer_location, dict)
            else printer_location
        )
        self.printer_properties = (
            PrinterProperties.from_dict(printer_properties)
            if isinstance(printer_properties, dict)
            else printer_properties
        )
        self.connection_parameters = (
            ConnectionParameters.from_dict(connection_parameters)
            if isinstance(connection_parameters, dict)
            else connection_parameters
        )
        self.backup_printer_properties = (
            BackupPrinterProperties.from_dict(backup_printer_properties)
            if isinstance(backup_printer_properties, dict)
            else backup_printer_properties
        )
        self.temp_file_location = temp_file_location


class SharePointDeviceProperties(Dictable):
    """Device properties for SharePoint device type.

    Attributes:
        file_location (FileLocation): File location options
        file_system (FileSystem): File system options
        iam_id (str, optional): IAM service ID\
    """

    _FROM_DICT_MAP = {
        "file_location": FileLocation.from_dict,
        "file_system": FileSystem.from_dict,
    }

    def __init__(
        self,
        file_location: FileLocation | dict,
        file_system: FileSystem | dict,
        iam_id: str | None = None,
    ):
        self.file_location = (
            FileLocation.from_dict(file_location)
            if isinstance(file_location, dict)
            else file_location
        )
        self.file_system = (
            FileSystem.from_dict(file_system)
            if isinstance(file_system, dict)
            else file_system
        )
        self.iam_id = iam_id


class S3Credentials(Dictable):
    """Connection parameters for S3 devices.

    Attributes:
        region (str): The AWS region where the S3 bucket is hosted,
            e.g. 'us-east-1'.
        access_key_id (str): The AWS Access Key ID used for authenticating
            API requests
        secret_access_key (str, optional): The AWS Secret Access Key paired with
            Access Key ID for secure authentication
    """

    def __init__(
        self,
        region: str,
        access_key_id: str,
        secret_access_key: str | None = None,
    ):
        self.region = region
        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key

    @classmethod
    def from_dict_or_none(cls, *args, **kwargs):
        # REST may return empty object for s3_credentials. We want to treat it
        # the same way as if the field was not set
        if not args and not kwargs:
            return None
        return cls(*args, **kwargs)


class S3DeviceProperties(SharePointDeviceProperties):
    """Device properties for S3 device type.

    Attributes:
        file_location (FileLocation): File location options
        file_system (FileSystem): File system options
        s3_credentials (S3Credentials): Connection parameters for S3
    """

    _FROM_DICT_MAP = {
        "file_location": FileLocation.from_dict,
        "file_system": FileSystem.from_dict,
        "s3_credentials": S3Credentials.from_dict_or_none,
    }

    def __init__(
        self,
        file_location: FileLocation | dict,
        file_system: FileSystem | dict,
        s3_credentials: S3Credentials | dict | None = None,
    ):
        self.file_location = (
            FileLocation.from_dict(file_location)
            if isinstance(file_location, dict)
            else file_location
        )
        self.file_system = (
            FileSystem.from_dict(file_system)
            if isinstance(file_system, dict)
            else file_system
        )
        # if not present, REST may return empty object
        if not s3_credentials:
            self.s3_credentials = None
        self.s3_credentials = (
            S3Credentials.from_dict(s3_credentials)
            if isinstance(s3_credentials, dict)
            else s3_credentials
        )



from dataclasses import dataclass, field
from enum import auto
from typing import TYPE_CHECKING, Optional

from mstrio.modeling.schema.helpers import SchemaObjectReference
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import camel_to_snake, delete_none_values, snake_to_camel

from .dynamic_date_time import DynamicDateTimeStructure
from .enums import Function, IsIndependent, NodeType
from .expression import Expression, ExpressionNode
from .parameters import AttributeElement, FunctionProperty, PredicateParameter, Variant

if TYPE_CHECKING:
    from mstrio.connection import Connection


@dataclass(kw_only=True)
class PredicateNode(ExpressionNode):
    """Base class for all predicate nodes.

    Attributes:
        predicate_id: String identifier that uniquely identifies this predicate
            within the context of its larger filter. Any non-empty string of
            printable characters can be used as a predicate.
        predicate_text: Text representation of a predicate
    """

    predicate_id: str | None = None
    predicate_text: str | None = None

    def to_dict(self, camel_case: bool = True) -> dict:
        result = {
            'type': self._TYPE.value,
            'predicate_id': self.predicate_id,
            'predicate_text': self.predicate_text,
            **self._get_node_data(),
        }

        result = delete_none_values(result, recursion=False)
        return snake_to_camel(result) if camel_case else result

    @classmethod
    def from_dict(
        cls,
        source: dict,
        connection: Optional['Connection'] = None,
        to_snake_case: bool = True,
    ) -> type['PredicateNode']:
        data = camel_to_snake(source) if to_snake_case else source.copy()
        # get a dict under predicate_tree attribute to unpack it
        predicate_tree = data.pop('predicate_tree')
        return super().from_dict(
            {**data, **predicate_tree}, connection, to_snake_case=False
        )

    def _get_node_data(self):
        """Return dictionary that contains data unique to particular predicate
        node that should be merged with data common to all predicate nodes in
        to_dict.
        """
        pass


@dataclass(kw_only=True)
class CustomExpressionPredicate(PredicateNode):
    """Specialized expression node that contains a custom expression predicate.
    This qualification contains a valid expression, usually created by the user,
    which can't be categorized into other types.

    The expression is represented by a list of tokens. The predicate has the
    predicate details which contains the tokens representing the expression.
    It's controlled by the showPredicates to show or hide.

    Attributes:
        expression: Generic specification for calculation, instance of
            Expression class
        node_property: A property used for persisting back-compatibility
            information for the filter editing. Usually it will not show.
    """

    _TYPE = NodeType.PREDICATE_CUSTOM
    _FROM_DICT_MAP = {
        **ExpressionNode._FROM_DICT_MAP,
        'expression': Expression,
    }

    expression: Expression
    node_property: int | None = None

    def _get_node_data(self):
        return {
            'node_property': self.node_property,
            'predicate_tree': {
                'expression': self.expression.to_dict(),
            },
        }


@dataclass(kw_only=True)
class MetricPredicate(PredicateNode):
    """Specialized expression node that contains a metric qualification
    predicate.

    This qualification selects tuples of elements at a specified level by
    comparing the value of a specified metric evaluated at the tuple with
    a value of an expression. The comparison operator used is also specified.
    However, there are multiple options for specifying the expression with which
    the metric is compared:

        - The client may specify a literal value.
        - The client may specify an object reference.
        - The client may ask to use a prompt to determine the value.
        - The client may supply an arbitrary metric style expression to evaluate
          to obtain the value.

    Attributes:
        function: Specify the comparison function that should be used to compare
            the value computed by the filter with an expression value.
        metric: The metric over whose values this qualification is evaluated.
        parameters: List of objects that correspond to the parameters for the
            comparison function.
            In nearly all cases the comparison function will take one extra
            parameter (in addition to the parameter referenced by the
            predicate's main object).
        level_type: Enumeration value indicating how the predicate's level
             should be determined.
             If there is a level it is usually taken to be the same level as is
             used in the display grid, or which was used to obtain the data.
             There is also an option to specify an absolute level as part of the
             predicate definition.
        level: Within an expression we sometimes need to state at which
            a computation is performed.
            In some cases the level is obtained by examining the execution
            environment. This object is used if the designer wishes to specify
            an explicit level. An explicit level is a list of attributes.
            Alternatively the client may defer selecting an attribute
            (or attributes) at design time by using a prompt instead.
            If a prompt is used it should be a choose-attributes prompt, and it
            may return any number of attributes (including zero).
        metric_function: Optional parameter that describes how the metric is
            used within the predicate. If this parameter is omitted it is
            treated as if it had been set to value.
        null_include: NullInclude is an integer parameter which specifies
            how null values are considered when evaluating the Rank function.
            Use 1 to place null values after the value list, use -1 to place
            null values before the value list or use 0 to apply the Null
            checking setting for Analytical Engine in VLDB properties. If this
            value is not set, 0 will be applied.
        break_by: The break by property of a metric qualification allows you to
            choose the attribute level at which to restart counting rank or
            percent values for a metric. This attribute level must be higher
            than or equal to the level of aggregation for the metric itself.
            break_by is a list of distinct attribute references. The list may
            also contain a choose-attribute prompt instead which can be answered
            with zero, one or more attributes. This value is only used when
            the metric_function property is not set to value.
        is_independent: Flag that indicates whether the metric will be evaluated
           independently of other parts of the larger filter:
           - If set to 1 (the default value, also used if this property is
               omitted) then this metric will be evaluated by itself.
           - If set to 0 then other parts of the larger filter will be applied
               when evaluating the metric.
    """

    class MetricFunction(AutoName):
        """Enumeration constant that describes how the metric is used
        within the predicate.
        """

        VALUE = auto()
        RANK_DESCEND = auto()
        RANK_ASCEND = auto()
        PERCENTILE_DESCEND = auto()
        PERCENTILE_ASCEND = auto()

    _TYPE = NodeType.PREDICATE_METRIC_QUALIFICATION
    _FROM_DICT_MAP = {
        **ExpressionNode._FROM_DICT_MAP,
        'function': Function,
        'parameters': [PredicateParameter.dispatch],
        'level': [SchemaObjectReference],
        'metric': SchemaObjectReference,
        'metric_function': MetricFunction,
        'break_by': [SchemaObjectReference],
        'is_independent': IsIndependent,
    }
    _ALLOW_NONE_ATTRIBUTES = ['parameters']

    function: Function
    metric: SchemaObjectReference
    level_type: str
    parameters: list[PredicateParameter] = field(default_factory=list)
    level: list[SchemaObjectReference] | None = None
    metric_function: MetricFunction | None = None
    null_include: int | None = None
    break_by: list[SchemaObjectReference] | None = None
    is_independent: IsIndependent = IsIndependent.YES

    def _get_node_data(self):
        result = {
            'predicate_tree': {
                'function': self.function.value,
                'parameters': [item.to_dict() for item in self.parameters],
                'level': (
                    [item.to_dict() for item in self.level] if self.level else None
                ),
                'level_type': self.level_type,
                'metric': self.metric.to_dict(),
                'metric_function': self.metric_function.value,
                'break_by': (
                    [item.to_dict() for item in self.break_by]
                    if self.break_by
                    else None
                ),
                'null_include': self.null_include,
                'is_independent': self.is_independent.value,
            }
        }

        return delete_none_values(
            result, recursion=True, whitelist_attributes=self._ALLOW_NONE_ATTRIBUTES
        )


@dataclass(kw_only=True)
class SetFromRelationshipPredicate(PredicateNode):
    """Specialized expression node that contains a relationship set
    qualification predicate. In words this predicate describes a qualification
    of the form "SET of WHERE".

    This node is used within filter expressions to represent a predicate whose
    value is determined by starting with an arbitrary filter expression at some
    natural level. This filter is then projected to a filter at a level
    specified within this predicate by using a relationship. The relationship
    might be the natural relationship inferred by the engine for the schema, the
    relationship given by an explicit table, or a relationship implied by a fact
    (i.e. the engine selects the best table from the set of tables where the
    given fact is implemented).

    This predicate is the only kind of filter predicate that may have a child
    expression.

    Attributes:
        level: Within an expression we sometimes need to state at which a
            computation is performed. In some cases the level is obtained by
            examining the execution environment. This object is used if the
            designer wishes to specify an explicit level.

            An explicit level is a list of attributes. Alternatively the client
            may defer selecting an attribute (or attributes) at design time by
            using a prompt instead. If a prompt is used it should be a
            choose-attributes prompt, and it may return any number of
            attributes (including zero).

        guide: An optional object that specifies how to relate the child filter
            with the output filter. There are three options, depending on the
            nature of the guide object:

            The object may be a logicalTable, in which case the filter is
            projected through this table. The object may be a fact, in which
            case the filter is projected via one of the tables on which the
            fact is defined. There might not be a guide object at all, in which
            case the filter is projected via the engine's usual join rules
            (using attribute relationships from the system hierarchy).

        is_independent: Flag that indicates whether this child filter will be
            considered independently of other parts of the larger filter:

            - If set to 1 (the default value, used if this property is omitted)
            then this filter will be evaluated by itself.

            - If set to 0 then other parts of the larger filter will be merged
            into this filter. Using this setting could change the value of a
            metric or relationship set qualification that appears within the
            child filter.

        children: List containing the optional root node of the child filter
            expression. We require that a relationship predicate always has a
            child filter is always given, but it is possible that the child
            filter is the empty filter. We model that situation by either
            omitting the children value or by using an empty list.

            This expression is also a filter expression, and it is manipulated
            in the same manner as the rest of the filter. The predicate ids
            that appear within this expression must have disjoint values from
            the predicate ids used in the larger filter. Thus, an individual
            node within this expression can be manipulated by treating it as a
            manipulation of the larger filter in the usual manner.

            We use a list so that this value matches the list value of the
            same name used in an operator node.
    """

    _TYPE = NodeType.PREDICATE_RELATIONSHIP
    _FROM_DICT_MAP = {
        **ExpressionNode._FROM_DICT_MAP,
        'level': [SchemaObjectReference],
        'guide': SchemaObjectReference,
        'is_independent': IsIndependent,
        'children': [ExpressionNode.dispatch],
    }

    level: list[SchemaObjectReference]
    guide: SchemaObjectReference | None = None
    is_independent: IsIndependent = IsIndependent.YES
    children: list[ExpressionNode] | None = None

    def _get_node_data(self):
        result = {
            'predicate_tree': {
                'level': [item.to_dict() for item in self.level],
                'guide': self.guide.to_dict() if self.guide else None,
                'is_independent': self.is_independent.value,
            },
            'children': (
                [item.to_dict() for item in self.children] if self.children else None
            ),
        }

        return delete_none_values(result, recursion=True)


@dataclass(kw_only=True)
class JointElementListPredicate(PredicateNode):
    """Specialized expression node that contains a joint element list predicate.

    This qualification represents a filter at the level of two or more
    attributes.
    The filter is defined by simply listing the tuples of elements that satisfy
    the filter.

    Attributes:
        level: The level of this filter represented as a list of attributes.
            The attributes must be different from each other, and their order
            is significant.

        tuples: List of tuples of elements.

            Each element of this list is a joint element. That is the members
            of this list are themselves lists of elements. The elements in
            the tuples must be listed in the same order as the attributes were
            listed in the level list. There is no need to list the attribute
            on each element.
    """

    _TYPE = NodeType.PREDICATE_JOINT_ELEMENT_LIST
    _FROM_DICT_MAP = {
        **ExpressionNode._FROM_DICT_MAP,
        'level': [SchemaObjectReference],
        'tuples': [[AttributeElement]],
    }

    level: list[SchemaObjectReference]
    tuples: list[list[AttributeElement]]

    def _get_node_data(self):
        return {
            'predicate_tree': {
                'level': [item.to_dict() for item in self.level],
                'tuples': [
                    [item.to_dict() for item in tuple_item]
                    for tuple_item in self.tuples
                ],
            },
        }


@dataclass(kw_only=True)
class ElementListPredicate(PredicateNode):
    """Specialized expression node that contains an element list predicate.

    This qualification selects elements of a specified attribute by either
    listing the elements that should be selected, or alternatively by listing
    the elements that should be excluded from the selection.

    Attributes:
        attribute: The attribute whose elements are selected by this
            qualification. Property is not needed if the predicate uses a
            prompt since in that case the attribute is specified by the prompt.

        elements: List of the elements in the predicate

        elements_prompt: A choose-elements prompt used to obtain the elements
            in the qualification. If this object is used then it will specify
            the attribute and elements for the qualification. The prompt will
            specify an attribute from which the elements should be selected.

        function: Specify the function that should be used in this predicate.

            Although this property is a function identifier, there are only two
            functions that make sense for this predicate:

            - IN (the filter accepts any element in the list, default value)
            - NOT_IN (the filter accepts any element not included in the list)
    """

    class ElementListFunction(AutoName):
        """Enumeration constant that describes what functions can be used
        within ElementListPredicate
        """

        IN = auto()
        NOT_IN = auto()

    _TYPE = NodeType.PREDICATE_ELEMENT_LIST
    _FROM_DICT_MAP = {
        **ExpressionNode._FROM_DICT_MAP,
        'attribute': SchemaObjectReference,
        'elements': [AttributeElement],
        'elements_prompt': SchemaObjectReference,
        'function': ElementListFunction,
    }

    attribute: SchemaObjectReference | None = None
    elements: list[AttributeElement] | None = None
    elements_prompt: SchemaObjectReference | None = None
    function: ElementListFunction = ElementListFunction.IN

    def _get_node_data(self):
        result = {
            'predicate_tree': {
                'attribute': self.attribute.to_dict() if self.attribute else None,
                'elements': (
                    [item.to_dict() for item in self.elements]
                    if self.elements
                    else None
                ),
                'elements_prompt': (
                    self.elements_prompt.to_dict() if self.elements_prompt else None
                ),
                'function': self.function.value,
            }
        }

        return delete_none_values(result, recursion=True)


@dataclass(kw_only=True)
class AttributeFormPredicate(PredicateNode):
    """Specialized expression node that contains an attribute form qualification
    predicate.

    This qualification selects elements of a specified attribute by comparing
    the value of
    a specified form of the element with an expression. The comparison operator
    used is also specified. However, there are three options for specifying the
    expression with which it is compared:

    - The client may specify a literal value.
    - The client may ask to use a prompt to determine the value.
    - The client may supply an arbitrary metric style expression to evaluate to
    obtain the value.

    Attributes:
        function: Specify the comparison function that should be used to
            compare the value computed by the filter with an expression value.

        parameters: list of objects that correspond to the parameters for the
            comparison function. In nearly all cases the comparison function
            will take one extra parameter (in addition to the parameter
            referenced by the predicate's main object).

            This value is a list because there are a few functions which take
            a different number of extra parameters. For example isNull does not
            need any additional parameters; notBetween needs two additional
            parameters.

        attribute: The attribute whose elements are selected by this
            qualification

        form: The attribute form whose value is used to specify the
            qualification

        data_locale: Optional data locale used to select values of the form.
            The format should follow the IETF BCP 47 language tag, for example:
            "en-US".

            Some forms of some attributes may be configured to store
            representation in multiple locales. If a predicate qualifies on a
            form that has been translated in this manner it is likely that the
            elements selected by the predicate will be different for different
            locales. But we do not want the meaning of a filter to vary based
            on the locale preferences of the user executing the filter. So this
            property is used to specify the data locale to be used for the
            qualification. When set it will override the user's personal data
            locale preference.

            Retrieving a filter only returns this field if the form is
            multilingual. If the form is multilingual but no particular locale
            is set, an empty string will be returned. When updating or creating
            a filter, again a check will be made to see if the form is capable
            of being multilingual. If not, users inputs are ignored. If the
            form is indeed multilingual and the field is omitted from the
            input, the locale will default to the warehouse data locale of the
            user.
    """

    _TYPE = NodeType.PREDICATE_FORM_QUALIFICATION
    _FROM_DICT_MAP = {
        **ExpressionNode._FROM_DICT_MAP,
        'parameters': [PredicateParameter.dispatch],
        'attribute': SchemaObjectReference,
        'form': SchemaObjectReference,
        'function': Function,
    }
    _ALLOW_NONE_ATTRIBUTES = ['parameters']

    function: Function
    attribute: SchemaObjectReference
    form: SchemaObjectReference
    # in swagger is required, but can be empty list
    parameters: list[PredicateParameter] = field(default_factory=list)
    data_locale: str | None = None

    def _get_node_data(self):
        result = {
            'predicate_tree': {
                'function': self.function.value,
                'attribute': self.attribute.to_dict(),
                'form': self.form.to_dict(),
                'parameters': [item.to_dict() for item in self.parameters],
                'data_locale': self.data_locale,
            }
        }

        return delete_none_values(
            result, recursion=True, whitelist_attributes=self._ALLOW_NONE_ATTRIBUTES
        )


@dataclass(kw_only=True)
class FilterQualificationPredicate(PredicateNode):
    """Specialized expression node that contains a filter qualification
    predicate.

    This node is used within filter expressions to represent a predicate whose
    value is determined by using an external filter object. At execution time
    the engine will typically act as if the specified filter had been included
    directly. By using shared filter objects, the client can build complex
    filter expressions to utilise common subexpressions.

    A filter object may not reference itself, or any filter that depends on
    itself. If such a filter is saved to the metadata any calculation that uses
    the filter will fail.

    Attributes:
        filter: Reference to the child filter that should be included in this
            predicate.

        is_independent: Flag that indicates whether this child filter will be
            considered independently of other parts of the larger filter:

            - If set to 1 (the default value, used if this property is omitted)
            then this filter will be evaluated by itself.

            - If set to 0 then other parts of the larger filter will be merged
            into this filter. Using this setting could change the value of a
            metric or relationship set qualification that appears within the
            child filter.
    """

    _TYPE = NodeType.PREDICATE_FILTER_QUALIFICATION
    _FROM_DICT_MAP = {
        **ExpressionNode._FROM_DICT_MAP,
        'filter': SchemaObjectReference,
        'is_independent': IsIndependent,
    }

    filter: SchemaObjectReference
    is_independent: IsIndependent = IsIndependent.YES

    def _get_node_data(self):
        return {
            'predicate_tree': {
                'filter': self.filter.to_dict(),
                'is_independent': self.is_independent.value,
            },
        }


@dataclass(kw_only=True)
class ReportQualificationPredicate(PredicateNode):
    """Specialized expression node that contains a report qualification
    predicate.

    This node is used within filter expressions to represent a predicate
    whose value is determined by using a report as filter. That is the engine
    will evaluate a report, and determine the rows that it covers. These filter
    will consist of these rows.

    Attributes:
        report: Reference to the child report that should be included in this
            predicate.
    """

    _TYPE = NodeType.PREDICATE_REPORT_QUALIFICATION
    _FROM_DICT_MAP = {
        **ExpressionNode._FROM_DICT_MAP,
        'report': SchemaObjectReference,
    }

    report: SchemaObjectReference

    def _get_node_data(self):
        return {
            'predicate_tree': {
                'report': self.report.to_dict(),
            },
        }


@dataclass(kw_only=True)
class PromptPredicate(PredicateNode):
    """Specialized expression node for a predicate whose value is prompted at
    execution time.

    This predicate is used for various kinds of expression prompts. At execution
    time the end user in effect enters an entire predicate (or an even more
    complex expression).
    The answer is then substituted into the filter in place of this node.

    Attributes:
        prompt: reference to a prompt, whose value is prompted at execution
    """

    _TYPE = NodeType.PREDICATE_PROMPT_QUALIFICATION
    _FROM_DICT_MAP = {
        **ExpressionNode._FROM_DICT_MAP,
        'prompt': SchemaObjectReference,
    }

    prompt: SchemaObjectReference

    def _get_node_data(self):
        return {
            'predicate_tree': {
                'prompt': self.prompt.to_dict(),
            },
        }


@dataclass(kw_only=True)
class Operator(ExpressionNode):
    """An expression node whose value is obtained by applying a specified
    function to a list of child expressions. Most internal nodes within
    an expression are operator nodes.

    Attributes:

        children: The list of child expression that combined at this operator.
            If there are multiple child expressions then their order
            is significant. In a few cases (depending on the function) it is
            valid for an operator node to have no children, in which case this
            member property may be omitted.

        function: Enumeration constant that identifies all the built-in
            functions. This value will always be provided when the service
            generates the object. When the client generates the object it may
            use this field by itself to specify the function. This value by
            itself is insufficient to identify the function if it is set to
            CUSTOM or THIRD_PARTY, in which case the custom_function value must
            be provided. This property is required in the predicate model but
            can be replaced by 'function_prompt' in the expression model.

        function_properties: The list of additional properties associated with
            the function. This is most useful when the function is 'RANK', as
            we'll need the "by_value" and "ascending" properties. This field is
            only used in a custom expression tree.

        function_prompt: The prompt used by this function. This field is only
            used in a custom expression tree.

        level_type: Enumeration value indicating how the predicate's level
            should be determined.

            If there is a level it is usually taken to be the same level as is
            used in the display grid, or which was used to obtain the data.
            There is also an option to specify an absolute level as part of the
            predicate definition.

            This field is only used in a custom expression tree.

        level: This field is only used in a custom expression tree.

        node_property: A property used for persisting back-compatibility
            information for the filter editing. Usually it will not show. If
            it's displayed in the response, you should consider include it with
            the same value in the same node in the request body when you're
            performing the PUT/POST/PATCH.

            If you don't put it back in the request, there can be some behavior
            difference in Filter Expression Editors of other Strategy One
            apps.
    """

    class LevelType(AutoName):
        """Enumeration value indicating how the predicate's level should be
        determined
        """

        NONE = auto()
        METRIC_LEVEL = auto()
        GRID_LEVEL = auto()
        EXPLICIT_LEVEL = auto()

    _TYPE = NodeType.OPERATOR
    _FROM_DICT_MAP = {
        **ExpressionNode._FROM_DICT_MAP,
        'children': [ExpressionNode.dispatch],
        'function': Function,
        'function_properties': [FunctionProperty],
        'function_prompt': SchemaObjectReference,
        'level_type': LevelType,
        'level': [SchemaObjectReference],
        'node_property': int,
        'custom_function': SchemaObjectReference,
    }

    children: list[ExpressionNode] | None = None
    function: Function | None = None
    function_properties: list[FunctionProperty] | None = None
    function_prompt: SchemaObjectReference | None = None
    level_type: LevelType | None = None
    level: list[SchemaObjectReference] | None = None
    node_property: int | None = None
    custom_function: SchemaObjectReference | None = None


@dataclass(kw_only=True)
class ObjectReference(ExpressionNode):
    """An expression node whose value is a reference to some other object in
    the metadata.

    The meaning of the reference depends on the context. If the referenced
    object is an object that is based on an expression (say a filter or a
    metric) then it usually means we inline the expression from the object.
    But in other cases (say a fact) it means that on execution the object
    reference is replaced with a value of the fact in the context of execution.

    Although unusual, it is possible that an object reference node might not
    refer to an object. If that happens this target property is usually omitted
    (but might be present but use 00000000000000000000000000000000 as its
    object_id).

    The object could be a prompt to select the target object at runtime.
    Typically, it is not acceptable for the prompt to return more than
    one answer.

    Attributes:
        target: object being referenced

        is_independent: Flag that indicates whether this child filter will be
            considered independently of other parts of the larger filter:

            - If set to 1 (the default value, used if this property is omitted)
            then this filter will be evaluated by itself.

            - If set to 0 then other parts of the larger filter will be merged
            into this filter. Using this setting could change the value of a
            metric or relationship set qualification that appears within the
            child filter.

        substitute_function_type: Specify how to merge multiple answers if the
            predicate's value is obtained via a prompt.
            This property is only used when filter's value is a prompt. If the
            end user selects multiple answers to the prompt, the platform will
            replace this node by a Boolean operator, used to combine the
            answers. This property determines which operator it uses. By
            default, it uses the parent Boolean operator (or AND if this
            predicate is the root of the filter). Set this property to override
            the default behavior.
    """

    class SubstituteFunctionType(AutoName):
        """Enumeration constant that specify how to merge multiple answers
        if the predicate's value is obtained via a prompt.
        """

        AND = auto()
        OR = auto()

    _TYPE = NodeType.OBJECT_REFERENCE
    _FROM_DICT_MAP = {
        **ExpressionNode._FROM_DICT_MAP,
        'target': SchemaObjectReference,
        'is_independent': IsIndependent,
        'substitute_function_type': SubstituteFunctionType,
    }

    target: SchemaObjectReference | None = None
    is_independent: IsIndependent = IsIndependent.YES
    substitute_function_type: SubstituteFunctionType | None = None


@dataclass(kw_only=True)
class ColumnReference(ExpressionNode):
    """n expression node whose value is some column from a table implied
    by the expression's context.

    This kind of node is used when we wish to define an expression over
    the columns of a table or potentially over more than one table.
    The reference specifies, the column of interest, by name alone. So the
    reference is to a selected column on the implied table. It is not
    a reference to a specific column on a specific table. Defining a column
    reference in this manner allows the architect to specify an expression
    that can be usefully applied to more than one table.

    Attributes:
        column_name: The name of the column, as it appears within the RDBMS.

        object_id: A globally unique identifier used to distinguish between
            metadata objects within the same project. It is possible for two
            metadata objects in different projects to have the same Object id.
    """

    _TYPE = NodeType.COLUMN_REFERENCE
    column_name: str
    object_id: str | None = None


@dataclass(kw_only=True)
class Constant(ExpressionNode):
    """An expression node that contains a literal value. The value was
    fixed when the expression node was first specified. This kind of node
    is only used for simple values (integers, floating point numbers and text
    strings).

    Alternatively, instead of the variant the node could contain a choose-value
    prompt reference.

    Attributes:
        variant: constant value specified as Variant
        prompt: reference to a prompt
    """

    _TYPE = NodeType.CONSTANT
    _FROM_DICT_MAP = {
        **ExpressionNode._FROM_DICT_MAP,
        'variant': Variant,
        'prompt': SchemaObjectReference,
    }

    variant: Variant | None = None
    prompt: SchemaObjectReference | None = None


@dataclass(kw_only=True)
class DynamicDateTime(ExpressionNode):
    """An expression node that represents Dynamic Date Time.

    Attributes:
        value (DynamicDateTimeStructure): an object that represents
            date and/or time.
    """

    _FROM_DICT_MAP = {
        **ExpressionNode._FROM_DICT_MAP,
        'value': DynamicDateTimeStructure,
    }
    _TYPE = NodeType.DYNAMIC_DATE_TIME
    value: DynamicDateTimeStructure


@dataclass(kw_only=True)
class ExpressionFormShortcut(ExpressionNode):
    """An expression node that represents shortcut to an expression form.

    Attributes:

        attribute (SchemaObjectReference): Information about an object
            referenced within the specification of another object. An object
            reference typically contains only enough fields to uniquely
            identify the referenced objects.

        form (SchemaObjectReference): Information about an object referenced
            within the specification of another object. An object reference
            typically contains only enough fields to uniquely identify the
            referenced objects.

        data_locale (str): Optional data locale used to select values of the
            form. The format should follow the IETF BCP 47 language tag, for
            example: "en-US".

            Some forms of some attributes may be configured to store
            representation in multiple locales. If a predicate qualifies on a
            form that has been translated in this manner it is likely that the
            elements selected by the predicate will be different for different
            locales. But we do not want the meaning of a filter to vary based
            on the locale preferences of the user executing the filter. So this
            property is used to specify the data locale to be used for the
            qualification. When set it will override the user's personal data
            locale preference.
    """

    _TYPE = NodeType.FORM_SHORTCUT
    _FROM_DICT_MAP = {
        **ExpressionNode._FROM_DICT_MAP,
        'attribute': SchemaObjectReference,
        'form': SchemaObjectReference,
    }

    attribute: SchemaObjectReference
    form: SchemaObjectReference
    data_locale: str | None = None


@dataclass(kw_only=True)
class ExpressionRelationship(ExpressionNode):
    """An expression node that represents expression relationship.

    Attributes:
        level: Within an expression we sometimes need to state at which a
            computation is performed. In some cases the level is obtained by
            examining the execution environment. This object is used if the
            designer wishes to specify an explicit level.

            An explicit level is a list of attributes. Alternatively the client
            may defer selecting an attribute (or attributes) at design time by
            using a prompt instead. If a prompt is used it should be a
            choose-attributes prompt, and it may return any number of
            attributes (including zero).

        guide: An optional object that specifies how to relate the child filter
            with the output filter. There are three options, depending on the
            nature of the guide object:

            - The object may be a logicalTable, in which case the filter is
            projected through this table.

            - The object may be a fact, in which case the filter is projected
            via one of the tables on which the fact is defined.

            - There might not be a guide object at all, in which case the
            filter is projected via the engine's usual join rules (using
            attribute relationships from the system hierarchy).

        is_independent: Flag that indicates whether this child filter will be
            considered independently of other parts of the larger filter:

            - If set to 1 (the default value, used if this property is omitted)
            then this filter will be evaluated by itself.

            - If set to 0 then other parts of the larger filter will be merged
            into this filter. Using this setting could change the value of a
            metric or relationship set qualification that appears within the
            child filter.

        children: List containing the optional root node of the child filter
            expression. We require that a relationship predicate always has a
            child filter is always given, but it is possible that the child
            filter is the empty filter. We model that situation by either
            omitting the children value or by using an empty list.

            This expression is also a filter expression, and it is manipulated
            in the same manner as the rest of the filter. The predicate ids
            that appear within this expression must have disjoint values from
            the predicate ids used in the larger filter. Thus, an individual
            node within this expression can be manipulated by treating it as a
            manipulation of the larger filter in the usual manner.

            We use a list so that this value matches the list value of the
            same name used in an operator node.
    """

    _TYPE = NodeType.RELATIONSHIP
    _FROM_DICT_MAP = {
        **ExpressionNode._FROM_DICT_MAP,
        'level': [SchemaObjectReference],
        'guide': SchemaObjectReference,
        'is_independent': IsIndependent,
        'children': [ExpressionNode.dispatch],
    }

    level: list[SchemaObjectReference]
    guide: SchemaObjectReference | None = None
    is_independent: IsIndependent = IsIndependent.YES
    children: list[ExpressionNode] | None = None

    def to_dict(self, camel_case: bool = True) -> dict:
        data = super().to_dict()
        result = {'type': data.pop('type'), 'relationship': data}

        return result if camel_case else camel_to_snake(result)

    @classmethod
    def from_dict(
        cls, source: dict, connection: 'Connection' = None, to_snake_case: bool = True
    ) -> 'ExpressionRelationship':
        data = source.get('relationship')
        return super().from_dict(data, connection, to_snake_case)


@dataclass(kw_only=True)
class BandingPredicate(PredicateNode):
    """Base class for banding predicates

    Attributes:
        level: level at which the computation will be performed

        metric: metric for which this banding will be applied

        band_metric_function: function used to slice metric values into bands

        band_names: list of band names, a band name cannot contain "#;" or "#,"
    """

    class BandMetricFunction(AutoName):
        """Enumeration constant used to specify function used to slice
        metric values into bands
        """

        VALUE = auto()
        RANK_DESCEND = auto()
        PERCENTILE_DESCEND = auto()

    _FROM_DICT_MAP = {
        **ExpressionNode._FROM_DICT_MAP,
        'level': [SchemaObjectReference],
        'metric': SchemaObjectReference,
        'band_metric_function': BandMetricFunction,
    }
    level: list[SchemaObjectReference]
    metric: SchemaObjectReference
    band_metric_function: BandMetricFunction

    def _get_banding_common_data(self):
        return {
            'predicate_tree': {
                'level': [item.to_dict() for item in self.level],
                'metric': self.metric.to_dict(),
                'band_metric_function': self.band_metric_function.value,
            },
        }


@dataclass(kw_only=True)
class BandingSizePredicate(BandingPredicate):
    """The Band size type of banding qualification slices a range of metric
    values into a number of bands that appear as rows on a report.

    You define the range by setting the start at and stop at values.
    You also set the step size, which is the size of each band.

    Attributes:
        start: the start of the range
        stop: the end of the range
        size: the size of each band
    """

    _TYPE = NodeType.PREDICATE_BANDING_SIZE
    _FROM_DICT_MAP = {
        **BandingPredicate._FROM_DICT_MAP,
        'start': Variant,
        'stop': Variant,
        'size': Variant,
    }

    start: Variant
    stop: Variant
    size: Variant
    band_names: list[str] | None = None

    def _get_node_data(self):
        result = self._get_banding_common_data()
        result['predicate_tree'].update(
            {
                'start': self.start.to_dict(),
                'stop': self.stop.to_dict(),
                'size': self.size.to_dict(),
                'band_names': self.band_names,
            }
        )

        return result


@dataclass(kw_only=True)
class BandingCountPredicate(BandingPredicate):
    """The Band count type of banding qualification slices a range of metric
    values into a number of equal bands that appear as rows on a report.

    You define the range by setting the start at and stop at values.
    You also set the band count, which is the number of bands to use.

    Attributes:
        start: the start of the range
        stop: the end of the range
        count: number of bands
    """

    _TYPE = NodeType.PREDICATE_BANDING_COUNT
    _FROM_DICT_MAP = {
        **BandingPredicate._FROM_DICT_MAP,
        'start': Variant,
        'stop': Variant,
        'count': Variant,
    }

    start: Variant
    stop: Variant
    count: Variant
    band_names: list[str] | None = None

    def _get_node_data(self):
        result = self._get_banding_common_data()
        result['predicate_tree'].update(
            {
                'start': self.start.to_dict(),
                'stop': self.stop.to_dict(),
                'count': self.count.to_dict(),
                'band_names': self.band_names,
            }
        )

        return result


@dataclass(kw_only=True)
class BandingPointsPredicate(BandingPredicate):
    """The Banding points type of banding qualification slices a range of metric
    values into a number of bands that appear as rows on a report.

    You manually define each band, which allows you to produce bands of varying
    sizes.

    Attributes:
        points: list of number used to slice metric values into bands
    """

    _TYPE = NodeType.PREDICATE_BANDING_POINTS
    points: list[float]
    band_names: list[str] | None = None

    def _get_node_data(self):
        result = self._get_banding_common_data()
        result['predicate_tree'].update(
            {
                'points': self.points,
                'band_names': self.band_names,
            }
        )

        return result


@dataclass(kw_only=True)
class BandingDistinctPredicate(BandingPredicate):
    """The Band for each distinct metric value type of banding qualification
    creates a separate band for each value calculated by the metric. The bands
    appear as rows on a report. This type of banding qualification directly uses
    the results of a metric as bands.

    It is very useful with metrics that already contain the logic needed to
    calculate sequential band numbers. Such metrics use mathematical formulas,
    NTile functions, Band functions, or Case functions.
    """

    _TYPE = NodeType.PREDICATE_BANDING_DISTINCT

    def _get_node_data(self):
        return self._get_banding_common_data()


# Dictionary mapping node types to class representing that node
NODE_TYPE_TO_CLASS_MAP = {
    NodeType.COLUMN_REFERENCE: ColumnReference,
    NodeType.CONSTANT: Constant,
    NodeType.DYNAMIC_DATE_TIME: DynamicDateTime,
    NodeType.FORM_SHORTCUT: ExpressionFormShortcut,
    NodeType.OBJECT_REFERENCE: ObjectReference,
    NodeType.OPERATOR: Operator,
    NodeType.PREDICATE_CUSTOM: CustomExpressionPredicate,
    NodeType.PREDICATE_PROMPT_QUALIFICATION: PromptPredicate,
    NodeType.PREDICATE_METRIC_QUALIFICATION: MetricPredicate,
    NodeType.PREDICATE_RELATIONSHIP: SetFromRelationshipPredicate,
    NodeType.PREDICATE_JOINT_ELEMENT_LIST: JointElementListPredicate,
    NodeType.PREDICATE_ELEMENT_LIST: ElementListPredicate,
    NodeType.PREDICATE_FORM_QUALIFICATION: AttributeFormPredicate,
    NodeType.PREDICATE_FILTER_QUALIFICATION: FilterQualificationPredicate,
    NodeType.PREDICATE_REPORT_QUALIFICATION: ReportQualificationPredicate,
    NodeType.PREDICATE_BANDING_SIZE: BandingSizePredicate,
    NodeType.PREDICATE_BANDING_COUNT: BandingCountPredicate,
    NodeType.PREDICATE_BANDING_POINTS: BandingPointsPredicate,
    NodeType.PREDICATE_BANDING_DISTINCT: BandingDistinctPredicate,
    NodeType.RELATIONSHIP: ExpressionRelationship,
}



import logging
from collections import defaultdict
from enum import auto
from typing import TYPE_CHECKING

from mstrio import config
from mstrio.api import devices
from mstrio.distribution_services.device.device_properties import (
    AndroidDeviceProperties,
    EmailDeviceProperties,
    FileDeviceProperties,
    FtpDeviceProperties,
    IOSDeviceProperties,
    PrinterDeviceProperties,
    SharePointDeviceProperties,
    S3DeviceProperties,
)
from mstrio.distribution_services.transmitter import Transmitter
from mstrio.types import ObjectTypes
from mstrio.users_and_groups import User
from mstrio.utils.entity import DeleteMixin, Entity
from mstrio.utils.enum_helper import AutoName, get_enum_val
from mstrio.utils.helper import (
    Dictable,
    delete_none_values,
    fetch_objects,
    get_args_from_func,
    get_default_args_from_func,
    get_objects_id,
)
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.version_helper import class_version_handler

if TYPE_CHECKING:
    from mstrio.connection import Connection

logger = logging.getLogger(__name__)


class DeviceType(AutoName):
    EMAIL = auto()
    FTP = auto()
    FILE = auto()
    PRINTER = auto()
    IPHONE = auto()
    IPAD = auto()
    ANDROID = auto()
    ONEDRIVE = auto()
    SHAREPOINT = auto()
    S3 = auto()
    ALL = auto()
    UNSUPPORTED = auto()


def list_devices(
    connection: 'Connection', to_dictionary: bool = False, limit: int = None, **filters
) -> list['Device'] | list[dict]:
    """Get list of Device objects or dicts. Optionally filter the
    devices by specifying filters.

    Args:
        connection: Strategy One connection object returned by
            `connection.Connection()`
        to_dictionary: If True returns dict, by default (False) returns
            Device objects.
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'date_created', 'date_modified', 'device_type', 'transmitter',
            'device_properties']

    Examples:
        >>> list_devices(connection, name='device_name')
    """
    return Device._list_devices(
        connection=connection, to_dictionary=to_dictionary, limit=limit, **filters
    )


@class_version_handler('11.3.0100')
class Device(Entity, DeleteMixin):
    """Devices are Distribution Services components that specify the format
     and transmission process of subscribed reports and documents.
     They are instances of transmitters that contain specific settings
     specific to a user's environments.

    Attributes:
        name: name of the device
        id: identifier of the Device
        description: description of the device
        device_type: type of the Device, DeviceType Enum
        transmitter: information of Transmitter attached
        device_properties: properties of the device,
            different for every device type
        ancestors: List of ancestor folders
        type: Object type. Enum
        subtype: Object subtype
        ext_type: Object extended type.
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        owner: User object that is the owner
        version: Object version ID
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
    """

    _OBJECT_TYPE = ObjectTypes.SUBSCRIPTION_DEVICE
    _DEVICE_TYPE_MAP = {
        'android': AndroidDeviceProperties,
        'email': EmailDeviceProperties,
        'file': FileDeviceProperties,
        'ftp': FtpDeviceProperties,
        'ipad': IOSDeviceProperties,
        'iphone': IOSDeviceProperties,
        'printer': PrinterDeviceProperties,
        'onedrive': SharePointDeviceProperties,
        'sharepoint': SharePointDeviceProperties,
        's3': S3DeviceProperties,
    }

    @staticmethod
    def _map_device_properties(source, connection):
        device_type = next(iter(source))
        device_properties = source[device_type]

        return Device._DEVICE_TYPE_MAP[device_type].from_dict(
            device_properties, connection
        )

    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'device_type': DeviceType,
        'owner': User.from_dict,
        'transmitter': Transmitter.from_dict,
        'device_properties': _map_device_properties,
    }
    _API_GETTERS = {
        (
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
            'comments',
        ): objects_processors.get_info,
        (
            'id',
            'name',
            'description',
            'device_type',
            'transmitter',
            'device_properties',
        ): devices.get_device,
    }
    _API_DELETE = staticmethod(devices.delete_device)
    _API_PATCH: dict = {
        ('name', 'description', 'device_properties'): (devices.update_device, 'put'),
        'comments': (objects_processors.update, 'partial_put'),
    }
    _PATCH_PATH_TYPES = {
        **Entity._PATCH_PATH_TYPES,
        'device_properties': dict,
    }

    def __init__(self, connection: 'Connection', name: str = None, id: str = None):
        """Initialize Device object."""

        if id is None and name is None:
            raise ValueError(
                "Please specify either 'id' or 'name' parameter in the constructor."
            )

        if id is None:
            objects_info = Device._list_devices(
                connection=connection, name=name, to_dictionary=True
            )
            if objects_info:
                object_info, object_info['connection'] = objects_info[0], connection
                self._init_variables(**object_info)
            else:
                raise ValueError(f"There is no Device: '{name}'")
        else:
            super().__init__(connection=connection, object_id=id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        device_type = kwargs.get('device_type')
        self._device_type = DeviceType(device_type) if device_type else None
        self._transmitter = (
            Transmitter.from_dict(kwargs.get('transmitter'), self.connection)
            if kwargs.get('transmitter')
            else None
        )
        device_properties = kwargs.get('device_properties')
        self.device_properties = (
            self._DEVICE_TYPE_MAP[device_type].from_dict(
                device_properties[device_type], self.connection
            )
            if device_properties and device_type
            else None
        )

    @classmethod
    def create(
        cls,
        connection: 'Connection',
        name: str,
        device_type: DeviceType | str,
        transmitter: Transmitter | str,
        device_properties: dict | Dictable,
        description: str = None,
    ) -> 'Device':
        """Create a new device.

        Args:
            connection: Strategy One connection object returned by
                `connection.Connection()`
            name: device object name
            device_type: type of the device
            transmitter: Transmitter object
            description: device object description
            device_properties: properties of the device
        Returns:
            Device object.
        """
        device_type = get_enum_val(device_type, DeviceType)
        device_properties = (
            device_properties.to_dict()
            if isinstance(device_properties, Dictable)
            else device_properties
        )
        transmitter_id = get_objects_id(transmitter, Transmitter)
        body = {
            'name': name,
            'description': description,
            'deviceType': device_type,
            'transmitter': {'id': transmitter_id},
            'deviceProperties': {device_type: device_properties},
        }
        body = delete_none_values(body, recursion=True)
        response = devices.create_device(connection, body).json()
        if config.verbose:
            logger.info(
                f"Successfully created device named: '{name}' with ID: '"
                f"{response['id']}'."
            )
        return cls.from_dict(source=response, connection=connection)

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        device_properties: Dictable | dict | None = None,
        comments: str | None = None,
    ):
        """Alter the device object properties.

        Args:
            name: device object name
            description: device object description
            device_properties: properties of the device, specific for each
                device type
            comments: long description of the device object
        """
        device_properties = device_properties or self.device_properties
        device_properties = (
            device_properties.to_dict()
            if isinstance(device_properties, Dictable)
            else device_properties
        )
        func = self.alter
        args = get_args_from_func(func)
        defaults = get_default_args_from_func(func)
        defaults_dict = dict(zip(args[-len(defaults) :], defaults)) if defaults else {}
        local = locals()
        properties = defaultdict(dict)
        for property_key in defaults_dict.keys():
            if property_key == 'device_properties':
                properties[property_key][self.device_type.value] = device_properties
            elif local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    def update_properties(self) -> None:
        """Save compatible local changes of the object attributes to the
        I-Server.

        Raises:
            requests.HTTPError: if I-Server raises exception
        """
        changes = {k: v[1] for k, v in self._altered_properties.items()}
        if 'device_properties' in changes:
            device_properties = changes['device_properties']
            del changes['device_properties']
        else:
            device_properties = self.device_properties
        changes['device_properties'] = {}
        changes['device_properties'][self.device_type.value] = (
            device_properties.to_dict()
            if not isinstance(device_properties, dict)
            else device_properties
        )
        self._alter_properties(**changes)
        self._altered_properties.clear()

    @classmethod
    def _list_devices(
        cls,
        connection: 'Connection',
        to_dictionary: bool = False,
        limit: int = None,
        **filters,
    ) -> list['Device'] | list[dict]:
        objects = fetch_objects(
            connection=connection,
            api=devices.get_devices,
            dict_unpack_value='devices',
            limit=limit,
            filters=filters,
        )
        if to_dictionary:
            return objects
        return [cls.from_dict(source=obj, connection=connection) for obj in objects]

    @property
    def device_type(self):
        return self._device_type

    @property
    def transmitter(self):
        return self._transmitter



"""
Import user or user group from csv.
Automate administrative tasks such as resetting password, assigning security
filters.

1. Connect to the environment using data from workstation
2. Create users from CSV file
3. Get first user group from the list of user groups
4. Set view access to the user group for all users created from CSV file
5. Reset password for all newly created users and make sure they have to change
   it at the next login
6. Import user group structure from CSV file
7. Apply security filter with name 'security_filter' to the user group with name
   'AI Users'
"""

import csv
from mstrio.connection import get_connection
from mstrio.helpers import AggregatedRights
from mstrio.modeling.security_filter.security_filter import list_security_filters
from mstrio.users_and_groups.user import User, create_users_from_csv
from mstrio.users_and_groups.user_group import UserGroup, list_user_groups

# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, project_name='MicroStrategy Tutorial')

# set correct user group access in tree view for user groups
# assumptions about CSV - csv is created from User.to_csv_from_list with at least
# two columns: username and full_name
user_list = create_users_from_csv(conn, 'users.csv')
user_group = list_user_groups(conn)[0]
user_group.acl_add(
    right=AggregatedRights.VIEW,
    trustees=user_list,
    denied=False,
    inheritable=True,
    propagate_to_children=True,
)

# reset the password for those users and check-in the option
# "User must change password at the next login"
for user in user_list:
    user.alter(password='new_password', require_new_password=True)

# import user group structure from CSV file
# assumptions - you have a prepared list CSV of two columns:
# user_name , user_group_name with headers added
with open('tree_structure.csv', encoding='utf-8') as structure:
    user_rows = csv.DictReader(structure)
    for row in user_rows:
        user_name = row['user_name']
        user_group_name = row['user_group_name']
        user = User(conn, name=user_name)
        user_group = UserGroup(conn, name=user_group_name)
        user_group.add_users(user)

# assigning correct security filter (the same name) for the user group
user_group = UserGroup(conn, name='AI Users')
security_filter = list_security_filters(conn, name='security_filter')[0]
user_group.apply_security_filter(security_filter)



# flake8: noqa
from typing import Union

from .user import User, create_users_from_csv, list_users
from .user_connections import UserConnections
from .user_group import UserGroup, list_user_groups

# isort: split

# those import has to be below import of `User` to avoid circular imports
from .contact import Contact, ContactAddress, ContactDeliveryType, list_contacts
from .contact_group import (
    ContactGroup,
    ContactGroupMember,
    ContactGroupMemberType,
    list_contact_groups,
)

UserOrGroup = Union[str, User, UserGroup]



import logging
from typing import TYPE_CHECKING

from mstrio import config
from mstrio.api import documents, library, objects
from mstrio.api.schedules import get_contents_schedule
from mstrio.helpers import IServerError
from mstrio.distribution_services.schedule import Schedule
from mstrio.users_and_groups import User, UserGroup, UserOrGroup
from mstrio.utils.helper import get_response_json
from mstrio.utils.version_helper import method_version_handler
from mstrio.object_management.library_shortcut import LibraryShortcut

if TYPE_CHECKING:
    from mstrio.utils.entity import Entity


REPORT_PROPERTIES_PROPERTY_SET_ID = "70A27C6E239911D5BF2200B0D02A21E0"
ALLOW_HTML_EXECUTION_PROPERTY_INDEX = 12


logger = logging.getLogger(__name__)


class LibraryMixin:
    """Mixin class for methods related to publishing objects to Library.
    Must be used with Entity or its subclasses. The class must:
        - populate its `instance_id` field,
        - be able to fetch its `recipients` field.

    """

    def _validate_user_or_group(self: "Entity", recipient_id: str) -> str | None:
        valid = False
        for cls in (User, UserGroup):
            try:
                cls(self.connection, id=recipient_id)
                valid = True
                break
            except IServerError:
                continue
        if not valid:
            if config.verbose:
                logger.info(
                    f'{recipient_id} is not a valid value for User or UserGroup ID'
                )
            return None
        return recipient_id

    def _recipients_to_ids(self, recipients: list[UserOrGroup | str]) -> list[str]:
        """Convert a list of str, User and UserGroup objects to a list of
        recipient (User or UserGroup) IDs and validate the ID strings.

        Args:
            recipients (list[UserOrGroup | str]): List of User or UserGroup
            references (objects or strs).
        Returns:
            list[str]: List of user IDs.
        """
        recipient_ids = []
        for recipient in recipients:
            if isinstance(recipient, str):
                if self._validate_user_or_group(recipient):
                    recipient_ids.append(recipient)
            elif isinstance(recipient, (User, UserGroup)):
                recipient_ids.append(recipient.id)
            else:
                raise ValueError(
                    'Please provide a list of User, UserGroup or str elements.'
                )
        return recipient_ids

    def publish(
        self: "Entity",
        recipients: UserOrGroup | str | list[UserOrGroup | str] | None = None,
    ):
        """Publish the document for authenticated user. If `recipients`
        parameter is specified publishes the document for the given users.

        Args:
            recipients(UserOrGroup | list[UserOrGroup | str] | str, optional):
                list of users or user groups to publish the document to
                (can be a list of IDs and User and UserGroup elements)
        """
        if not isinstance(recipients, list) and recipients is not None:
            recipients = [recipients]

        if recipients is None:
            recipients = [self.connection.user_id]
        recipients = self._recipients_to_ids(recipients)

        body = {'id': self.id, 'recipients': recipients}
        library.publish_document(self.connection, body)
        self.fetch(attr='recipients')

    def unpublish(
        self: "Entity",
        recipients: UserOrGroup | str | list[UserOrGroup | str] | None = None,
    ):
        """Unpublish the document for all users it was previously published to.
        If `recipients` parameter is specified unpublishes the document for the
        given users or user groups.

        Args:
            recipients(UserOrGroup | list[UserOrGroup | str] | str, optional):
                list of users or user groups to unpublish the document from
                (can be a list of IDs and User and UserGroup elements)
        """

        if recipients is None:
            library.unpublish_document(self.connection, id=self.id)
        else:
            if not isinstance(recipients, list):
                recipients = [recipients]
            recipients = self._recipients_to_ids(recipients)

            for recipient_id in recipients:
                library.unpublish_document_for_user(
                    self.connection, document_id=self.id, recipient_id=recipient_id
                )

        self.fetch(attr='recipients')

    @method_version_handler('11.3.0600')
    def list_available_schedules(
        self: "Entity",
        to_dictionary: bool = False,
        name: str | None = None,
        limit: int | None = None,
    ) -> list["Schedule"] | list[dict]:
        """Get a list of schedules available for the object instance.

        Args:
            to_dictionary (bool, optional): If True returns a list of
                dictionaries, otherwise returns a list of Schedules.
                False by default.
            name (str, optional): If specified, filters the schedules to those
                with the specified substring in their name.
            limit (int, optional): If specified, limits the number of schedules
                returned.

        Returns:
            List of Schedule objects or list of dictionaries.
        """
        schedules_list: list[dict] = (
            get_response_json(
                get_contents_schedule(
                    connection=self.connection,
                    project_id=self.connection.project_id,
                    body={'id': self.id, 'type': 'document'},
                )
            )
        ).get('schedules')

        if name:
            schedules_list = [
                schedule for schedule in schedules_list if name in schedule['name']
            ]
        if limit is not None:
            schedules_list = schedules_list[:limit]

        if to_dictionary:
            return schedules_list
        else:
            return [
                Schedule.from_dict(connection=self.connection, source=schedule_id)
                for schedule_id in schedules_list
            ]

    def share_to(self: "Entity", users: UserOrGroup | list[UserOrGroup]):
        """Shares the document to the listed users' libraries.

        Args:
            users(UserOrGroup | list[UserOrGroup]): list of users or user
                groups to publish the document to (can be a list of IDs or a
                list of User and UserGroup elements).
        """
        self.publish(users)

    def is_html_js_execution_enabled(self: "Entity") -> bool | None:
        """Check whether HTML and JavaScript execution is enabled
        for the document.

        Returns:
            bool: True if HTML and JavaScript execution is enabled,
                False otherwise.
        """
        res = get_response_json(
            objects.get_property_set(
                self.connection,
                id=self.id,
                obj_type=self._OBJECT_TYPE.value,
                property_set_id=REPORT_PROPERTIES_PROPERTY_SET_ID,
            )
        )
        prop_in_list = [
            prop for prop in res if prop['id'] == ALLOW_HTML_EXECUTION_PROPERTY_INDEX
        ]
        try:
            prop = bool(prop_in_list[0]['value'])
        except IndexError:
            raise AttributeError(
                "This object does not support HTML/JS execution property"
            )

        return prop

    def set_html_js_execution_enabled(self: "Entity", enabled: bool) -> None:
        """Enable or disable HTML and JavaScript execution for the document.

        Args:
            enabled (bool): True to enable HTML and JavaScript execution,
                False to disable.
        """

        body = [
            {
                "properties": [
                    {"value": int(enabled), "id": ALLOW_HTML_EXECUTION_PROPERTY_INDEX}
                ],
                "id": REPORT_PROPERTIES_PROPERTY_SET_ID,
            }
        ]
        objects.update_property_set(
            self.connection, id=self.id, obj_type=self._OBJECT_TYPE.value, body=body
        )

    def get_library_shortcut(self: "Entity") -> "LibraryShortcut":
        response: dict = documents.get_document_shortcut(
            connection=self.connection,
            document_id=self.id,
            instance_id=self.instance_id,
        ).json()
        return LibraryShortcut.from_dict(connection=self.connection, source=response)



"""This is the demo script to show how to manage license. Its basic goal
is to present what can be done with this module and to ease its usage.
"""

from time import sleep

from mstrio.connection import get_connection
from mstrio.server import ContactInformation, InstallationUse, License

conn = get_connection(workstationData)

# Get license object
license = License(connection=conn)

# Print activation information
print(license.activation_info)

# Get license details
license_details = license.list_details()

# Get all license properties
license_properties = license.list_properties()

# Get all license products
license_products = license.list_products(module='all')
# Get licensed products under 'intelligence_module'
license_products_intelligence = license.list_products()
# Get licensed products under 'identity_module'
license_products_identity = license.list_products(module='identity_module')
# Get licensed products beginning with 'Server'
license_products_server = license.list_products(module='all', name='Server')
# Get licensed products with id 7
license_products_id_7 = license.list_products(module='all', product_id=7)

# Print license status
print(license.status)
# Deactivate license
license.deactivate()
# Print license status
print(license.status)


NEW_LICENSE_KEY = $new_license_key
# Before registering new license key, you can check its entitlements
new_entitlements = license.get_new_key_entitlements(key=NEW_LICENSE_KEY)
# Print new entitlements
print(new_entitlements)
# Register new license key
license.register_new_key(NEW_LICENSE_KEY)
# Print license key
print(license.key)

# To load new license key, you need to restart the server
# After restarting the server, you can check the license key and status
print(license.key)
print(license.status)

# To activate new license key, you need to provide contact information first
# Contact information

DEPARTMENT = $department
FIRST_NAME = $first_name
LAST_NAME = $last_name
ADDRESS = $address
POSTAL = $postal
COUNTRY = $country
JOB_TITLE = $job_title
CITY = $city
PHONE = $phone
EMAIL = $email
CONFIRM_EMAIL = $confirm_email
COMPANY = $company
customer_contact_info = ContactInformation(department=DEPARTMENT,
                                            first_name=FIRST_NAME,
                                            last_name=LAST_NAME,
                                            address=ADDRESS,
                                            postal=POSTAL,
                                            country=COUNTRY,
                                            job_title=JOB_TITLE,
                                            city=CITY,
                                            phone=PHONE,
                                            email=EMAIL,
                                            confirm_email=CONFIRM_EMAIL,
                                            company=COMPANY)

LOCATION = $location

license.update_activation_request_information(location=LOCATION,
                                                installation_use=InstallationUse.PRODUCTION,
                                                customer_contact=customer_contact_info)

# Download activation xml file
ACTIVATION_XML_FILE = $activation_xml_file
license.get_activation_xml_file(save_path=ACTIVATION_XML_FILE)

# Now to obtain activation code you need to upload activation xml file into
# Strategy Licensing site and get activation code

ACTIVATION_CODE = $activation_code
# Activate license with activation code
license.activate(activation_code=ACTIVATION_CODE)
# Print activation information
print(license.activation_info)
# Print license status
print(license.status)

# License compliance check run and get
license.run_compliance_check()
# Wait for compliance check to finish
sleep(5)
# Get compliance check information
license.get_compliance_check()
# Alternatively, you can run and get compliance check in one step
license.run_and_get_compliance_check()
# Print compliance time
print(license.compliance_time)
# Print compliance result
print(license.in_compliance)
# Now you can check the compliance status for each product
print(license.list_products())

# License audit
license.run_audit()
# Wait for audit to finish
sleep(2)
# Get audit information
license.get_audit()
# Alternatively, you can run and get audit in one step
license.run_and_get_audit()
# Print audit time
print(license.audit_time)

# Export standard audit results to csv
license.export_audit_to_csv(file_path="audit_standard.csv")

# If there was a product with compliance issues, you can selectively fetch
# users' privileges for that product and export to csv semi-detailed
# audit results
for product in license.list_products(module='all'):
    if not product.in_compliance:
        product.fetch_users_privileges()
license.export_audit_to_csv(file_path="audit_semi_detailed.csv")

# Fetch all products and their users' privileges
# Depending on total number of users in the environment, this can take several
# minutes to complete
license.fetch_all_products_users_privileges()
# Export detailed audit results to csv
license.export_audit_to_csv(file_path="audit_detailed.csv")

# Export to .json
license.export_audit_to_dataframe(file_path = "detailed_audit.json")

# Export to .pkl
license.export_audit_to_dataframe(file_path = "detailed_audit.pkl")

# Export to df
df = license.export_audit_to_dataframe()
print(df.head())
